@@ -1,4 +1,4 @@
-/* $Id: anchor.c,v 1.11 2002/09/05 15:43:21 ukai Exp $ */
+/* $Id: anchor.c,v 1.26 2003/02/11 10:36:17 ukai Exp $ */
 #include "fm.h"
 #include "myctype.h"
 #include "regex.h"
@@ -7,7 +7,7 @@
 
 AnchorList *
 putAnchor(AnchorList *al, char *url, char *target, Anchor **anchor_return,
-	  char *referer, int line, int pos)
+	  char *referer, char *title, unsigned char key, int line, int pos)
 {
     int n, i, j;
     Anchor *a;
@@ -44,6 +44,9 @@ putAnchor(AnchorList *al, char *url, char *target, Anchor **anchor_return,
     a->url = url;
     a->target = target;
     a->referer = referer;
+    a->title = title;
+    a->accesskey = key;
+    a->slave = FALSE;
     a->start = bp;
     a->end = bp;
     al->nanchor++;
@@ -54,11 +57,12 @@ putAnchor(AnchorList *al, char *url, char *target, Anchor **anchor_return,
 
 
 Anchor *
-registerHref(Buffer *buf, char *url, char *target, char *referer, int line,
-	     int pos)
+registerHref(Buffer *buf, char *url, char *target, char *referer, char *title,
+	     unsigned char key, int line, int pos)
 {
     Anchor *a;
-    buf->href = putAnchor(buf->href, url, target, &a, referer, line, pos);
+    buf->href = putAnchor(buf->href, url, target, &a, referer, title, key,
+			  line, pos);
     return a;
 }
 
@@ -66,15 +70,17 @@ Anchor *
 registerName(Buffer *buf, char *url, int line, int pos)
 {
     Anchor *a;
-    buf->name = putAnchor(buf->name, url, NULL, &a, NULL, line, pos);
+    buf->name = putAnchor(buf->name, url, NULL, &a, NULL, NULL, '\0', line,
+			  pos);
     return a;
 }
 
 Anchor *
-registerImg(Buffer *buf, char *url, int line, int pos)
+registerImg(Buffer *buf, char *url, char *title, int line, int pos)
 {
     Anchor *a;
-    buf->img = putAnchor(buf->img, url, NULL, &a, NULL, line, pos);
+    buf->img = putAnchor(buf->img, url, NULL, &a, NULL, title, '\0', line,
+			 pos);
     return a;
 }
 
@@ -88,8 +94,8 @@ registerForm(Buffer *buf, FormList *flist, struct parsed_tag *tag, int line,
     fi = formList_addInput(flist, tag);
     if (fi == NULL)
 	return NULL;
-    buf->formitem = putAnchor(buf->formitem,
-			      (char *)fi, flist->target, &a, NULL, line, pos);
+    buf->formitem = putAnchor(buf->formitem, (char *)fi, flist->target, &a,
+			      NULL, NULL, '\0', line, pos);
     return a;
 }
 
@@ -187,21 +193,26 @@ _put_anchor_news(Buffer *buf, char *p1, char *p2, int line, int pos)
 {
     Str tmp;
 
-    p1++;
-    if (*(p2 - 1) == '>')
-	p2--;
-    tmp = Strnew_size(sizeof("news:") - 1 + (p2 - p1));
-    Strcat_charp_n(tmp, "news:", sizeof("news:") - 1);
-    Strcat_charp_n(tmp, p1, p2 - p1);
-    return registerHref(buf, tmp->ptr, NULL, NO_REFERER, line, pos);
+    if (*p1 == '<') {
+	p1++;
+	if (*(p2 - 1) == '>')
+	    p2--;
+    }
+    tmp = conv_str(Strnew_charp_n(p1, p2 - p1), InnerCode, buf->document_code);
+    tmp = Sprintf("news:%s", file_quote(tmp->ptr));
+    return registerHref(buf, tmp->ptr, NULL, NO_REFERER, NULL, '\0', line,
+			pos);
 }
 #endif				/* USE_NNTP */
 
 static Anchor *
 _put_anchor_all(Buffer *buf, char *p1, char *p2, int line, int pos)
 {
-    return registerHref(buf, allocStr(p1, p2 - p1), NULL, NO_REFERER, line,
-			pos);
+    Str tmp;
+
+    tmp = conv_str(Strnew_charp_n(p1, p2 - p1), InnerCode, buf->document_code);
+    return registerHref(buf, url_quote(tmp->ptr), NULL, NO_REFERER, NULL,
+			'\0', line, pos);
 }
 
 static void
@@ -283,7 +294,7 @@ reAnchorPos(Buffer *buf, Line *l, char *p1, char *p2,
 {
     Anchor *a;
     int spos, epos;
-    int i;
+    int i, hseq = -2;
 
     spos = p1 - l->lineBuf;
     epos = p2 - l->lineBuf;
@@ -291,12 +302,27 @@ reAnchorPos(Buffer *buf, Line *l, char *p1, char *p2,
 	if (l->propBuf[i] & (PE_ANCHOR | PE_FORM))
 	    return p2;
     }
-    a = anchorproc(buf, p1, p2, l->linenumber, p1 - l->lineBuf);
-    a->end.line = l->linenumber;
-    a->end.pos = epos;
-    a->hseq = -2;
-    for (i = a->start.pos; i < a->end.pos; i++)
+    for (i = spos; i < epos; i++)
 	l->propBuf[i] |= PE_ANCHOR;
+    while (1) {
+	a = anchorproc(buf, p1, p2, l->linenumber, spos);
+	a->hseq = hseq;
+	if (hseq == -2) {
+	    reseq_anchor(buf);
+	    hseq = a->hseq;
+	}
+	a->end.line = l->linenumber;
+	if (epos > l->len) {
+	    a->end.pos = l->len;
+	    spos = 0;
+	    epos -= l->len;
+	    l = l->next;
+	}
+	else {
+	    a->end.pos = epos;
+	    break;
+	}
+    }
     return p2;
 }
 
@@ -304,7 +330,6 @@ void
 reAnchorWord(Buffer *buf, Line *l, int spos, int epos)
 {
     reAnchorPos(buf, l, &l->lineBuf[spos], &l->lineBuf[epos], _put_anchor_all);
-    reseq_anchor(buf);
 }
 
 /* search regexp and register them as anchors */
@@ -314,7 +339,7 @@ reAnchorAny(Buffer *buf, char *re,
 	    Anchor *(*anchorproc) (Buffer *, char *, char *, int, int))
 {
     Line *l;
-    char *p, *p1, *p2;
+    char *p = NULL, *p1, *p2;
 
     if (re == NULL || *re == '\0') {
 	return NULL;
@@ -325,17 +350,22 @@ reAnchorAny(Buffer *buf, char *re,
     for (l = MarkAllPages ? buf->firstLine : buf->topLine; l != NULL &&
 	 (MarkAllPages || l->linenumber < buf->topLine->linenumber + LASTLINE);
 	 l = l->next) {
+	if (p && l->bpos)
+	    goto next_line;
 	p = l->lineBuf;
 	for (;;) {
-	    if (regexMatch(p, &l->lineBuf[l->len] - p, p == l->lineBuf) == 1) {
+	    if (regexMatch(p, &l->lineBuf[l->size] - p, p == l->lineBuf) == 1) {
 		matchedPosition(&p1, &p2);
 		p = reAnchorPos(buf, l, p1, p2, anchorproc);
 	    }
 	    else
 		break;
 	}
+      next_line:
+	if (MarkAllPages && l->next == NULL && buf->pagerSource &&
+	    !(buf->bufferprop & BP_CLOSE))
+	    getNextPage(buf, PagerMax);
     }
-    reseq_anchor(buf);
     return NULL;
 }
 
@@ -351,6 +381,63 @@ reAnchorNews(Buffer *buf, char *re)
 {
     return reAnchorAny(buf, re, _put_anchor_news);
 }
+
+char *
+reAnchorNewsheader(Buffer *buf)
+{
+    Line *l;
+    char *p, *p1, *p2;
+    static char *header_mid[] = {
+	"Message-Id:", "References:", "In-Reply-To:", NULL
+    };
+    static char *header_group[] = {
+	"Newsgroups:", NULL
+    };
+    char **header, **q;
+    int i, search = FALSE;
+
+    if (!buf || !buf->firstLine)
+	return NULL;
+    for (i = 0; i <= 1; i++) {
+	if (i == 0) {
+	    regexCompile("<[!-;=?-~]+@[a-zA-Z0-9\\.\\-_]+>", 1);
+	    header = header_mid;
+	}
+	else {
+	    regexCompile("[a-zA-Z0-9\\.\\-_]+", 1);
+	    header = header_group;
+	}
+	for (l = buf->firstLine; l != NULL && l->real_linenumber == 0;
+	     l = l->next) {
+	    if (l->bpos)
+		continue;
+	    p = l->lineBuf;
+	    if (!IS_SPACE(*p)) {
+		search = FALSE;
+		for (q = header; *q; q++) {
+		    if (!strncasecmp(p, *q, strlen(*q))) {
+			search = TRUE;
+			p = strchr(p, ':') + 1;
+			break;
+		    }
+		}
+	    }
+	    if (!search)
+		continue;
+	    for (;;) {
+		if (regexMatch(p, &l->lineBuf[l->size] - p, p == l->lineBuf)
+		    == 1) {
+		    matchedPosition(&p1, &p2);
+		    p = reAnchorPos(buf, l, p1, p2, _put_anchor_news);
+		}
+		else
+		    break;
+	    }
+	}
+    }
+    reseq_anchor(buf);
+    return NULL;
+}
 #endif				/* USE_NNTP */
 
 #define FIRST_MARKER_SIZE 30
@@ -471,23 +558,26 @@ addMultirowsImg(Buffer *buf, AnchorList *al)
 	    if (a_img.start.line == l->linenumber)
 		continue;
 	    pos = columnPos(l, col);
-	    a = registerImg(buf, a_img.url, l->linenumber, pos);
+	    a = registerImg(buf, a_img.url, a_img.title, l->linenumber, pos);
 	    a->hseq = -a_img.hseq;
+	    a->slave = TRUE;
 	    a->image = img;
 	    a->end.pos = pos + ecol - col;
 	    for (k = pos; k < a->end.pos; k++)
 		l->propBuf[k] |= PE_IMAGE;
 	    if (a_href.url) {
 		a = registerHref(buf, a_href.url, a_href.target,
-				 a_href.referer, l->linenumber, pos);
+				 a_href.referer, a_href.title,
+				 a_href.accesskey, l->linenumber, pos);
 		a->hseq = a_href.hseq;
+		a->slave = TRUE;
 		a->end.pos = pos + ecol - col;
 		for (k = pos; k < a->end.pos; k++)
 		    l->propBuf[k] |= PE_ANCHOR;
 	    }
 	    if (a_form.url) {
 		buf->formitem = putAnchor(buf->formitem, a_form.url,
-					  a_form.target, &a, NULL,
+					  a_form.target, &a, NULL, NULL, '\0',
 					  l->linenumber, pos);
 		a->hseq = a_form.hseq;
 		a->end.pos = pos + ecol - col;
@@ -542,8 +632,8 @@ addMultirowsForm(Buffer *buf, AnchorList *al)
 	    if (a_form.start.line == l->linenumber)
 		continue;
 	    buf->formitem = putAnchor(buf->formitem, a_form.url,
-				      a_form.target, &a, NULL, l->linenumber,
-				      pos);
+				      a_form.target, &a, NULL, NULL, '\0',
+				      l->linenumber, pos);
 	    a->hseq = a_form.hseq;
 	    a->y = a_form.y;
 	    a->end.pos = pos + ecol - col;
@@ -554,3 +644,173 @@ addMultirowsForm(Buffer *buf, AnchorList *al)
 	}
     }
 }
+
+char *
+getAnchorText(Buffer *buf, AnchorList *al, Anchor *a)
+{
+    int hseq, i;
+    Line *l;
+    Str tmp = NULL;
+    char *p, *ep;
+
+    if (!a || a->hseq < 0)
+	return NULL;
+    hseq = a->hseq;
+    l = buf->firstLine;
+    for (i = 0; i < al->nanchor; i++) {
+	a = &al->anchors[i];
+	if (a->hseq != hseq)
+	    continue;
+	for (; l; l = l->next) {
+	    if (l->linenumber == a->start.line)
+		break;
+	}
+	if (!l)
+	    break;
+	p = l->lineBuf + a->start.pos;
+	ep = l->lineBuf + a->end.pos;
+	for (; p < ep && IS_SPACE(*p); p++) ;
+	if (p == ep)
+	    continue;
+	if (!tmp)
+	    tmp = Strnew_size(ep - p);
+	else
+	    Strcat_char(tmp, ' ');
+	Strcat_charp_n(tmp, p, ep - p);
+    }
+    return tmp ? tmp->ptr : NULL;
+}
+
+Buffer *
+link_list_panel(Buffer *buf)
+{
+    LinkList *l;
+    AnchorList *al;
+    Anchor *a;
+    FormItemList *fi;
+    int i;
+    char *t, *u, *p;
+    ParsedURL pu;
+    Str tmp = Strnew_charp("<title>Link List</title>\
+<h1 align=center>Link List</h1>\n");
+
+    if (buf->bufferprop & BP_INTERNAL ||
+	(buf->linklist == NULL && buf->href == NULL && buf->img == NULL)) {
+	return NULL;
+    }
+
+    if (buf->linklist) {
+	Strcat_charp(tmp, "<hr><h2>Links</h2>\n<ol>\n");
+	for (l = buf->linklist; l; l = l->next) {
+	    if (l->url) {
+		parseURL2(l->url, &pu, baseURL(buf));
+		p = parsedURL2Str(&pu)->ptr;
+		u = html_quote(p);
+		if (DecodeURL)
+		    p = html_quote(url_unquote_conv(p, buf->document_code));
+		else
+		    p = u;
+	    }
+	    else
+		u = p = "";
+	    if (l->type == LINK_TYPE_REL)
+		t = " [Rel]";
+	    else if (l->type == LINK_TYPE_REV)
+		t = " [Rev]";
+	    else
+		t = "";
+	    t = Sprintf("%s%s\n", l->title ? l->title : "", t)->ptr;
+	    t = html_quote(t);
+	    Strcat_m_charp(tmp, "<li><a href=\"", u, "\">", t, "</a><br>", p,
+			   "\n", NULL);
+	}
+	Strcat_charp(tmp, "</ol>\n");
+    }
+
+    if (buf->href) {
+	Strcat_charp(tmp, "<hr><h2>Anchors</h2>\n<ol>\n");
+	al = buf->href;
+	for (i = 0; i < al->nanchor; i++) {
+	    a = &al->anchors[i];
+	    if (a->slave)
+		continue;
+	    parseURL2(a->url, &pu, baseURL(buf));
+	    p = parsedURL2Str(&pu)->ptr;
+	    u = html_quote(p);
+	    if (DecodeURL)
+		p = html_quote(url_unquote_conv(p, buf->document_code));
+	    else
+		p = u;
+	    t = getAnchorText(buf, al, a);
+	    t = t ? html_quote(t) : "";
+	    Strcat_m_charp(tmp, "<li><a href=\"", u, "\">", t, "</a><br>", p,
+			   "\n", NULL);
+	}
+	Strcat_charp(tmp, "</ol>\n");
+    }
+
+    if (buf->img) {
+	Strcat_charp(tmp, "<hr><h2>Images</h2>\n<ol>\n");
+	al = buf->img;
+	for (i = 0; i < al->nanchor; i++) {
+	    a = &al->anchors[i];
+	    if (a->slave)
+		continue;
+	    parseURL2(a->url, &pu, baseURL(buf));
+	    p = parsedURL2Str(&pu)->ptr;
+	    u = html_quote(p);
+	    if (DecodeURL)
+		p = html_quote(url_unquote_conv(p, buf->document_code));
+	    else
+		p = u;
+	    if (a->title && *a->title)
+		t = html_quote(a->title);
+	    else if (DecodeURL)
+		t = html_quote(url_unquote_conv(a->url, buf->document_code));
+	    else
+		t = html_quote(a->url);
+	    Strcat_m_charp(tmp, "<li><a href=\"", u, "\">", t, "</a><br>", p,
+			   "\n", NULL);
+	    a = retrieveAnchor(buf->formitem, a->start.line, a->start.pos);
+	    if (!a)
+		continue;
+	    fi = (FormItemList *)a->url;
+	    fi = fi->parent->item;
+	    if (fi->parent->method == FORM_METHOD_INTERNAL &&
+		!Strcmp_charp(fi->parent->action, "map") && fi->value) {
+		MapList *ml = searchMapList(buf, fi->value->ptr);
+		ListItem *mi;
+		MapArea *m;
+		if (!ml)
+		    continue;
+		Strcat_charp(tmp, "<br>\n<b>Image map</b>\n<ol>\n");
+		for (mi = ml->area->first; mi != NULL; mi = mi->next) {
+		    m = (MapArea *) mi->ptr;
+		    if (!m)
+			continue;
+		    parseURL2(m->url, &pu, baseURL(buf));
+		    p = parsedURL2Str(&pu)->ptr;
+		    u = html_quote(p);
+		    if (DecodeURL)
+			p = html_quote(url_unquote_conv(p,
+							buf->document_code));
+		    else
+			p = u;
+		    if (m->alt && *m->alt)
+			t = html_quote(m->alt);
+		    else if (DecodeURL)
+			t = html_quote(url_unquote_conv(m->url,
+							buf->document_code));
+		    else
+			t = html_quote(m->url);
+		    Strcat_m_charp(tmp, "<li><a href=\"", u, "\">", t,
+				   "</a><br>", p, "\n", NULL);
+		}
+		Strcat_charp(tmp, "</ol>\n");
+	    }
+	}
+	Strcat_charp(tmp, "</ol>\n");
+    }
+
+    return loadHTMLString(tmp);
+}
