@@ -1,8 +1,7 @@
-/* $Id: backend.c,v 1.9 2002/02/03 15:25:45 ukai Exp $ */
+/* $Id: backend.c,v 1.11 2003/01/23 18:37:20 ukai Exp $ */
 #include <stdio.h>
 #include <string.h>
 #include <sys/types.h>
-#include <ctype.h>
 #include "fm.h"
 #include "gc.h"
 #include "terms.h"
@@ -149,7 +148,7 @@ internal_get(char *url, int flag, FormList *request)
 		}
 		print_headers(buf, len);
 		printf("\n");
-		saveBuffer(buf, stdout);
+		saveBuffer(buf, stdout, TRUE);
 	    }
 	    else {
 		print_headers(buf, 0);
