@@ -1,4 +1,4 @@
-/* $Id: fm.h,v 1.67 2002/10/30 15:46:29 ukai Exp $ */
+/* $Id: fm.h,v 1.113 2003/02/18 15:43:24 ukai Exp $ */
 /* 
  * w3m: WWW wo Miru utility
  * 
@@ -71,12 +71,12 @@ void bzero(void *, int);
 /* 
  * Constants.
  */
-#define LINELEN	4096		/* Maximum line length */
+#define LINELEN	256		/* Initial line length */
 #define PAGER_MAX_LINE	10000	/* Maximum line kept as pager */
-#define FNLEN 80
 
 #ifdef USE_IMAGE
 #define MAX_IMAGE 1000
+#define MAX_IMAGE_SIZE 2048
 
 #define DEFAULT_PIXEL_PER_CHAR  7.0	/* arbitrary */
 #define DEFAULT_PIXEL_PER_LINE  14.0	/* arbitrary */
@@ -171,7 +171,6 @@ void bzero(void *, int);
 #define BP_NORMAL	0x0
 #define BP_PIPE		0x1
 #define BP_FRAME	0x2
-#define BP_SOURCE	0x4
 #define BP_INTERNAL	0x8
 #define BP_NO_URL	0x10
 #define BP_REDIRECTED   0x20
@@ -238,6 +237,7 @@ extern int REV_LB[];
 #define IMG_FLAG_UNLOADED	0
 #define IMG_FLAG_LOADED		1
 #define IMG_FLAG_ERROR		2
+#define IMG_FLAG_DONT_REMOVE	4
 
 /* 
  * Macros.
@@ -285,13 +285,14 @@ typedef unsigned char Linecolor;
 
 typedef struct _MapArea {
     char *url;
+    char *target;
     char *alt;
-#ifdef MENU_MAP
 #ifdef USE_IMAGE
     char shape;
     short *coords;
     int ncoords;
-#endif
+    short center_x;
+    short center_y;
 #endif
 } MapArea;
 
@@ -309,16 +310,19 @@ typedef struct _Line {
 #endif
     struct _Line *next;
     struct _Line *prev;
-    short len;
-    short width;
+    int len;
+    int width;
     long linenumber;		/* on buffer */
     long real_linenumber;	/* on file */
     unsigned short usrflags;
+    int size;
+    int bpos;
+    int bwidth;
 } Line;
 
 typedef struct {
     int line;
-    short pos;
+    int pos;
 } BufferPoint;
 
 #ifdef USE_IMAGE
@@ -354,9 +358,12 @@ typedef struct _anchor {
     char *url;
     char *target;
     char *referer;
+    char *title;
+    unsigned char accesskey;
     BufferPoint start;
     BufferPoint end;
     int hseq;
+    char slave;
     short y;
     short rows;
 #ifdef USE_IMAGE
@@ -380,6 +387,17 @@ typedef struct {
     int prevhseq;
 } HmarkerList;
 
+#define LINK_TYPE_NONE 0
+#define LINK_TYPE_REL  1
+#define LINK_TYPE_REV  2
+typedef struct _LinkList {
+    char *url;
+    char *title;	/* Next, Contents, ... */
+    char *ctype;	/* Content-Type */
+    char type;		/* Rel, Rev */
+    struct _LinkList *next;
+} LinkList;
+
 typedef struct _Buffer {
     char *filename;
     char *buffername;
@@ -395,18 +413,21 @@ typedef struct _Buffer {
     char *real_type;
     int allLine;
     short bufferprop;
-    short currentColumn;
+    int currentColumn;
     short cursorX;
     short cursorY;
-    short pos;
-    short visualpos;
+    int pos;
+    int visualpos;
     short rootX;
+    short rootY;
     short COLS;
+    short LINES;
     InputStream pagerSource;
     AnchorList *href;
     AnchorList *name;
     AnchorList *img;
     AnchorList *formitem;
+    LinkList *linklist;
     FormList *formlist;
     MapList *maplist;
     HmarkerList *hmarklist;
@@ -419,7 +440,6 @@ typedef struct _Buffer {
     struct frameset *frameset;
     struct frameset_queue *frameQ;
     int *clone;
-    size_t linelen;
     size_t trbyte;
     char check_url;
 #ifdef JP_CHARSET
@@ -437,9 +457,47 @@ typedef struct _Buffer {
     char *ssl_certificate;
 #endif
     char image_flag;
+    char image_loaded;
     char need_reshape;
+    Anchor *submit;
+    struct _BufferPos *undo;
+#ifdef USE_ALARM
+    struct _AlarmEvent *event;
+#endif
 } Buffer;
 
+typedef struct _BufferPos {
+    long top_linenumber;
+    long cur_linenumber;
+    int currentColumn;
+    int pos;
+    int bpos;
+    struct _BufferPos *next;
+    struct _BufferPos *prev;
+} BufferPos;
+
+typedef struct _TabBuffer {
+    struct _TabBuffer *nextTab;
+    struct _TabBuffer *prevTab;
+    Buffer *currentBuffer;
+    Buffer *firstBuffer;
+    short x1;
+    short x2;
+    short y;
+} TabBuffer;
+
+typedef struct _DownloadList {
+    pid_t pid;
+    char *url;
+    char *save;
+    char *lock;
+    clen_t size;
+    time_t time;
+    int ok;
+    struct _DownloadList *next;
+    struct _DownloadList *prev;
+} DownloadList;
+#define DOWNLOAD_LIST_TITLE "Download List Panel"
 
 #define COPY_BUFPOSITION(dstbuf, srcbuf) {\
  (dstbuf)->topLine = (srcbuf)->topLine; \
@@ -465,16 +523,16 @@ typedef struct _Buffer {
 
 #define FONTSTAT_SIZE 4
 
-#define INIT_BUFFER_WIDTH (COLS-1)
+#define _INIT_BUFFER_WIDTH (COLS - (showLineNum ? 6 : 1))
+#define INIT_BUFFER_WIDTH ((_INIT_BUFFER_WIDTH > 0) ? _INIT_BUFFER_WIDTH : 0)
+#define FOLD_BUFFER_WIDTH (FoldLine ? (INIT_BUFFER_WIDTH + 1) : -1)
 
 typedef struct {
     int pos;
     int len;
     int tlen;
     long flag;
-    Str anchor;
-    Str anchor_target;
-    short anchor_hseq;
+    Anchor anchor;
     Str img_alt;
     char fontstat[FONTSTAT_SIZE];
     short nobr_level;
@@ -493,12 +551,10 @@ struct readbuffer {
     long flag_stack[RB_STACK_SIZE];
     int flag_sp;
     int status;
-    Str ignore_tag;
+    unsigned char end_tag;
     short table_level;
     short nobr_level;
-    Str anchor;
-    Str anchor_target;
-    short anchor_hseq;
+    Anchor anchor;
     Str img_alt;
     char fontstat[FONTSTAT_SIZE];
     char fontstat_stack[FONT_STACK_SIZE][FONTSTAT_SIZE];
@@ -516,34 +572,33 @@ struct readbuffer {
 #define in_stand fontstat[2]
 
 #define RB_PRE		0x01
-#define RB_XMPMODE	0x02
-#define RB_LSTMODE	0x04
+#define RB_SCRIPT	0x02
+#define RB_STYLE	0x04
 #define RB_PLAIN	0x08
-#define RB_LEFT		0x80000
-#define RB_CENTER	0x10
-#define RB_RIGHT	0x20
-#define RB_ALIGN	(RB_LEFT| RB_CENTER | RB_RIGHT)
-#define RB_NOBR		0x40
-#define RB_P		0x80
-#define RB_PRE_INT	0x100
-#define RB_PREMODE	(RB_PRE | RB_PRE_INT)
-#define RB_SPECIAL	(RB_PRE|RB_XMPMODE|RB_LSTMODE|RB_PLAIN|RB_NOBR|RB_PRE_INT)
-#define RB_PLAINMODE	(RB_XMPMODE|RB_LSTMODE|RB_PLAIN)
-
-#define RB_IN_DT	0x200
-#define RB_INTXTA	0x400
-#define RB_INSELECT	0x800
-#define RB_IGNORE	0x1000
-#define RB_INSEL	0x2000
-#define RB_IGNORE_P	0x4000
-#define RB_TITLE	0x8000
-#define RB_NFLUSHED	0x10000
-#define RB_NOFRAMES	0x20000
-#define RB_INTABLE	0x40000
+#define RB_LEFT		0x10
+#define RB_CENTER	0x20
+#define RB_RIGHT	0x40
+#define RB_ALIGN	(RB_LEFT | RB_CENTER | RB_RIGHT)
+#define RB_NOBR		0x80
+#define RB_P		0x100
+#define RB_PRE_INT	0x200
+#define RB_IN_DT	0x400
+#define RB_INTXTA	0x800
+#define RB_INSELECT	0x1000
+#define RB_IGNORE_P	0x2000
+#define RB_TITLE	0x4000
+#define RB_NFLUSHED	0x8000
+#define RB_NOFRAMES	0x10000
+#define RB_INTABLE	0x20000
+#define RB_PREMODE	(RB_PRE | RB_PRE_INT | RB_SCRIPT | RB_STYLE | RB_PLAIN | RB_INTXTA)
+#define RB_SPECIAL	(RB_PRE | RB_PRE_INT | RB_SCRIPT | RB_STYLE | RB_PLAIN | RB_NOBR)
+#define RB_PLAIN_PRE	0x40000
 
 #ifdef FORMAT_NICE
-#define RB_FILL		0x200000
+#define RB_FILL		0x80000
 #endif				/* FORMAT_NICE */
+#define RB_DEL		0x100000
+#define RB_S		0x200000
 
 #define RB_GET_ALIGN(obuf) ((obuf)->flag&RB_ALIGN)
 #define RB_SET_ALIGN(obuf,align) {(obuf)->flag &= ~RB_ALIGN; (obuf)->flag |= (align); }
@@ -564,17 +619,17 @@ struct readbuffer {
 #define R_ST_DQUOTE 4		/* within double quote */
 #define R_ST_EQL    5		/* = */
 #define R_ST_AMP    6		/* within ampersand quote */
-#define R_ST_CMNT1  7		/* <!  */
-#define R_ST_CMNT2  8		/* <!- */
-#define R_ST_CMNT   9		/* within comment */
-#define R_ST_NCMNT1 10		/* comment - */
-#define R_ST_NCMNT2 11		/* comment -- */
-#define R_ST_NCMNT3 12		/* comment -- space */
-#define R_ST_IRRTAG 13		/* within irregular tag */
+#define R_ST_EOL    7		/* end of file */
+#define R_ST_CMNT1  8		/* <!  */
+#define R_ST_CMNT2  9		/* <!- */
+#define R_ST_CMNT   10		/* within comment */
+#define R_ST_NCMNT1 11		/* comment - */
+#define R_ST_NCMNT2 12		/* comment -- */
+#define R_ST_NCMNT3 13		/* comment -- space */
+#define R_ST_IRRTAG 14		/* within irregular tag */
+#define R_ST_VALUE  15		/* within tag attribule value */
 
-#define ST_IS_REAL_TAG(s)   ((s)==R_ST_TAG||(s)==R_ST_TAG0||(s)==R_ST_EQL)
-#define ST_IS_COMMENT(s)    ((s)>=R_ST_CMNT1)
-#define ST_IS_TAG(s)        ((s)!=R_ST_NORMAL&&(s)!=R_ST_AMP&&!ST_IS_COMMENT(s))
+#define ST_IS_REAL_TAG(s)   ((s)==R_ST_TAG||(s)==R_ST_TAG0||(s)==R_ST_EQL||(s)==R_ST_VALUE)
 
 /* is this '<' really means the beginning of a tag? */
 #define REALLY_THE_BEGINNING_OF_A_TAG(p) \
@@ -689,7 +744,8 @@ typedef struct http_request {
 #define TMPF_SRC	1
 #define TMPF_FRAME	2
 #define TMPF_CACHE	3
-#define MAX_TMPF_TYPE	4
+#define TMPF_COOKIE	4
+#define MAX_TMPF_TYPE	5
 
 #define set_no_proxy(domains) (NO_proxy_domains=make_domain_list(domains))
 
@@ -717,27 +773,46 @@ global char RenderFrame init(FALSE);
 global char TargetSelf init(FALSE);
 global char PermitSaveToPipe init(FALSE);
 global char DecodeCTE init(FALSE);
+global char PreserveTimestamp init(TRUE);
 global char ArgvIsURL init(FALSE);
 global char MetaRefresh init(FALSE);
 
 global char fmInitialized init(FALSE);
 global char QuietMessage init(FALSE);
+global char TrapSignal init(TRUE);
+#define TRAP_ON if (TrapSignal) { \
+    prevtrap = mySignal(SIGINT, KeyAbort); \
+    if (fmInitialized) \
+	term_cbreak(); \
+}
+#define TRAP_OFF if (TrapSignal) { \
+    if (fmInitialized) \
+	term_raw(); \
+    if (prevtrap) \
+	mySignal(SIGINT, prevtrap); \
+}
 
-extern char GlobalKeymap[];
-extern char EscKeymap[];
-extern char EscBKeymap[];
-extern char EscDKeymap[];
+extern unsigned char GlobalKeymap[];
+extern unsigned char EscKeymap[];
+extern unsigned char EscBKeymap[];
+extern unsigned char EscDKeymap[];
 #ifdef __EMX__
-extern char PcKeymap[];
+extern unsigned char PcKeymap[];
 #endif
 extern FuncList w3mFuncList[];
 
 global char *HTTP_proxy init(NULL);
+#ifdef USE_SSL
+global char *HTTPS_proxy init(NULL);
+#endif				/* USE_SSL */
 #ifdef USE_GOPHER
 global char *GOPHER_proxy init(NULL);
 #endif				/* USE_GOPHER */
 global char *FTP_proxy init(NULL);
 global ParsedURL HTTP_proxy_parsed;
+#ifdef USE_SSL
+global ParsedURL HTTPS_proxy_parsed;
+#endif				/* USE_SSL */
 #ifdef USE_GOPHER
 global ParsedURL GOPHER_proxy_parsed;
 #endif				/* USE_GOPHER */
@@ -753,8 +828,14 @@ extern int ai_family_order_table[3][3];	/* XXX */
 #endif				/* INET6 */
 global TextList *NO_proxy_domains;
 global char NoCache init(FALSE);
-global int Do_not_use_proxy init(FALSE);
+global char use_proxy init(TRUE);
+#define Do_not_use_proxy (!use_proxy)
 global int Do_not_use_ti_te init(FALSE);
+#ifdef USE_NNTP
+global char *NNTP_server init(NULL);
+global char *NNTP_mode init(NULL);
+global int MaxNewsMessage init(50);
+#endif
 
 global char *document_root init(NULL);
 global char *personal_document_root init(NULL);
@@ -762,8 +843,24 @@ global char *cgi_bin init(NULL);
 global char *index_file init(NULL);
 
 global char *CurrentDir;
+global int CurrentPid;
+/*
 global Buffer *Currentbuf;
 global Buffer *Firstbuf;
+*/
+global TabBuffer *CurrentTab;
+global TabBuffer *FirstTab;
+global TabBuffer *LastTab;
+global int open_tab_blank init(FALSE);
+global int open_tab_dl_list init(FALSE);
+global int close_tab_back init(FALSE);
+global int nTab;
+global int TabCols init(10);
+#define NO_TABBUFFER ((TabBuffer *)1)
+#define Currentbuf (CurrentTab->currentBuffer)
+#define Firstbuf (CurrentTab->firstBuffer)
+global DownloadList *FirstDL init(NULL);
+global DownloadList *LastDL init(NULL);
 global int CurrentKey;
 global char *CurrentKeyData;
 global char *CurrentCmdData;
@@ -818,6 +915,7 @@ global int nextpage_topline init(FALSE);
 global char *displayTitleTerm init(NULL);
 global int displayLink init(FALSE);
 global int displayLineInfo init(FALSE);
+global int DecodeURL init(FALSE);
 global int retryAsHttp init(TRUE);
 global int showLineNum init(FALSE);
 global int show_srch_str init(TRUE);
@@ -828,6 +926,7 @@ global int displayImage init(TRUE);
 global int autoImage init(TRUE);
 global int useExtImageViewer init(TRUE);
 global int maxLoadImage init(4);
+global int image_map_list init(TRUE);
 #else
 global int displayImage init(FALSE); /* XXX: emacs-w3m use display_image=off */
 #endif
@@ -841,7 +940,9 @@ global char *ExtBrowser init(DEF_EXT_BROWSER);
 global char *ExtBrowser2 init(NULL);
 global char *ExtBrowser3 init(NULL);
 global int BackgroundExtViewer init(TRUE);
+global int disable_secret_security_check init(FALSE);
 global char *passwd_file init(PASSWD_FILE);
+global char *pre_form_file init(PRE_FORM_FILE);
 global char *ftppasswd init(NULL);
 #ifdef FTPPASS_HOSTNAMEGEN
 global int ftppass_hostnamegen init(TRUE);
@@ -869,7 +970,9 @@ global int UseDictCommand init(FALSE);
 global char *DictCommand init("file:///$LIB/w3mdict" CGI_EXTENSION);
 #endif				/* USE_DICT */
 global int ignore_null_img_alt init(TRUE);
+global int displayInsDel init(TRUE);
 global int FoldTextarea init(FALSE);
+global int FoldLine init(FALSE);
 #define DEFAULT_URL_EMPTY	0
 #define DEFAULT_URL_CURRENT	1
 #define DEFAULT_URL_LINK	2
@@ -883,7 +986,7 @@ global char *migemo_command init(DEF_MIGEMO_COMMAND);
 #endif				/* USE_MIGEMO */
 
 global struct auth_cookie *Auth_cookie init(NULL);
-global char *Local_cookie init(NULL);
+global Str Local_cookie init(NULL);
 #ifdef USE_COOKIE
 global struct cookie *First_cookie init(NULL);
 #endif				/* USE_COOKIE */
@@ -926,17 +1029,43 @@ global char UseAutoDetect init(TRUE);
 #define url_quote_conv(x,c) url_quote(x)
 #endif				/* JP_CHARSET */
 #ifndef KANJI_SYMBOLS
-global int no_graphic_char init(FALSE);
+global char UseGraphicChar init(TRUE);
 extern char alt_rule[];
 #endif				/* not KANJI_SYMBOLS */
 extern char UseAltEntity;
-global char *rc_dir;
-global int rc_dir_is_tmp init(FALSE);
+global int no_rc_dir init(FALSE);
+global char *rc_dir init(NULL);
+global char *tmp_dir;
+global char *config_file init(NULL);
 
 #ifdef USE_MOUSE
 global int use_mouse init(TRUE);
 extern int mouseActive;
 global int reverse_mouse init(FALSE);
+global int relative_wheel_scroll init(FALSE);
+global int fixed_wheel_scroll_count init(5);
+global int relative_wheel_scroll_ratio init(30);
+typedef struct _MouseActionMap {
+    void (*func) ();
+    char *data;
+} MouseActionMap;
+typedef struct _MouseAction {
+    char *menu_str;
+    char *lastline_str;
+    int menu_width;
+    int lastline_width;
+    int in_action;
+    int cursorX;
+    int cursorY;
+    MouseActionMap default_map[3];
+    MouseActionMap anchor_map[3];
+    MouseActionMap active_map[3];
+    MouseActionMap tab_map[3];
+    MouseActionMap *menu_map[3];
+    MouseActionMap *lastline_map[3];
+} MouseAction;
+global MouseAction mouse_action;
+#define LIMIT_MOUSE_MENU 100
 #endif				/* USE_MOUSE */
 
 #ifdef USE_COOKIE
@@ -1015,8 +1144,14 @@ void w3m_exit(int i);
 #define AL_UNSET         0
 #define AL_EXPLICIT      1
 #define AL_IMPLICIT      2
-#define AL_IMPLICIT_DONE 3
-#define AL_IMPLICIT_ONCE 4
+#define AL_IMPLICIT_ONCE 3
+
+typedef struct _AlarmEvent {
+    int sec;
+    short status;
+    int cmd;
+    void *data;
+} AlarmEvent;
 #endif
 
 /* 
