@@ -1,4 +1,4 @@
-/* $Id: frame.c,v 1.16.4.1 2002/11/26 07:11:22 ukai Exp $ */
+/* $Id: frame.c,v 1.31 2003/01/29 17:10:39 ukai Exp $ */
 #include "fm.h"
 #include "parsetagx.h"
 #include "myctype.h"
@@ -97,7 +97,7 @@ newFrame(struct parsed_tag *tag, Buffer *buf)
     body->baseURL = baseURL(buf);
     if (tag) {
 	if (parsedtag_get_value(tag, ATTR_SRC, &p))
-	    body->url = url_quote_conv(p, buf->document_code);
+	    body->url = url_quote_conv(remove_space(p), buf->document_code);
 	if (parsedtag_get_value(tag, ATTR_NAME, &p) && *p != '_')
 	    body->name = url_quote_conv(p, buf->document_code);
     }
@@ -107,8 +107,6 @@ newFrame(struct parsed_tag *tag, Buffer *buf)
 static void
 unloadFrame(struct frame_body *b)
 {
-    if (b->source && b->flags & FB_TODELETE)
-	pushText(fileToDelete, b->source);
     b->attr = F_UNLOADED;
 }
 
@@ -177,7 +175,6 @@ copyFrame(struct frame_body *ob)
 
     rb = New(struct frame_body);
     bcopy((const void *)ob, (void *)rb, sizeof(struct frame_body));
-    rb->flags &= ~FB_TODELETE;
     return rb;
 }
 
@@ -316,21 +313,12 @@ resetFrameElement(union frameset_element *f_element,
 	f_body->attr = F_BODY;
 	f_body->name = f_name;
 	f_body->url = parsedURL2Str(&buf->currentURL)->ptr;
+	f_body->source = buf->sourcefile;
+	buf->sourcefile = NULL;
 	if (buf->mailcap_source) {
 	    f_body->source = buf->mailcap_source;
-	    f_body->flags |= FB_TODELETE;
 	    buf->mailcap_source = NULL;
 	}
-	else if (buf->real_scheme == SCM_LOCAL) {
-	    f_body->source = buf->sourcefile;
-	}
-	else {
-	    Str tmp = tmpfname(TMPF_FRAME, NULL);
-	    rename(buf->sourcefile, tmp->ptr);
-	    f_body->source = tmp->ptr;
-	    f_body->flags |= FB_TODELETE;
-	    buf->sourcefile = NULL;
-	}
 	f_body->type = buf->type;
 	f_body->referer = referer;
 	f_body->request = request;
@@ -345,7 +333,6 @@ frame_download_source(struct frame_body *b, ParsedURL *currentURL,
 {
     Buffer *buf;
     struct frameset *ret_frameset = NULL;
-    Str tmp;
     ParsedURL url;
 
     if (b == NULL || b->url == NULL || b->url[0] == '\0')
@@ -381,25 +368,13 @@ frame_download_source(struct frame_body *b, ParsedURL *currentURL,
 	return NULL;
     }
     b->url = parsedURL2Str(&buf->currentURL)->ptr;
-    b->source = buf->sourcefile;
     b->type = buf->type;
+    b->source = buf->sourcefile;
+    buf->sourcefile = NULL;
     if (buf->mailcap_source) {
 	b->source = buf->mailcap_source;
-	b->flags |= FB_TODELETE;
 	buf->mailcap_source = NULL;
     }
-    else if ((buf->real_scheme != SCM_LOCAL)
-#ifdef USE_IMAGE
-	     || (activeImage && !useExtImageViewer &&
-		 buf->real_type && !strncasecmp(buf->real_type, "image/", 6))
-#endif
-	) {
-	tmp = tmpfname(TMPF_FRAME, NULL);
-	rename(buf->sourcefile, tmp->ptr);
-	b->source = tmp->ptr;
-	b->flags |= FB_TODELETE;
-	buf->sourcefile = NULL;
-    }
     b->attr = F_BODY;
     if (buf->frameset) {
 	ret_frameset = buf->frameset;
@@ -412,14 +387,32 @@ frame_download_source(struct frame_body *b, ParsedURL *currentURL,
     return ret_frameset;
 }
 
+#define CASE_TABLE_TAG \
+	case HTML_TR:\
+	case HTML_N_TR:\
+	case HTML_TD:\
+	case HTML_N_TD:\
+	case HTML_TH:\
+	case HTML_N_TH:\
+	case HTML_THEAD:\
+	case HTML_N_THEAD:\
+	case HTML_TBODY:\
+	case HTML_N_TBODY:\
+	case HTML_TFOOT:\
+	case HTML_N_TFOOT:\
+	case HTML_COLGROUP:\
+	case HTML_N_COLGROUP:\
+	case HTML_COL
+
 static int
 createFrameFile(struct frameset *f, FILE * f1, Buffer *current, int level,
 		int force_reload)
 {
     int r, c, t_stack;
     URLFile f2;
+    char code;
 #ifdef JP_CHARSET
-    char code, ic, charset[2];
+    char charset[2];
 #endif				/* JP_CHARSET */
     char *d_target, *p_target, *s_target, *t_target;
     ParsedURL *currentURL, base;
@@ -431,15 +424,10 @@ createFrameFile(struct frameset *f, FILE * f1, Buffer *current, int level,
 
     if (level == 0) {
 	if (SETJMP(AbortLoading) != 0) {
-	    if (fmInitialized)
-		term_raw();
-	    signal(SIGINT, prevtrap);
+	    TRAP_OFF;
 	    return -1;
 	}
-	prevtrap = signal(SIGINT, KeyAbort);
-	if (fmInitialized)
-	    term_cbreak();
-
+	TRAP_ON;
 	f->name = "_top";
     }
 
@@ -467,8 +455,10 @@ createFrameFile(struct frameset *f, FILE * f1, Buffer *current, int level,
 	    struct frameset *f_frameset;
 	    int i = c + r * f->col;
 	    char *p = "";
+	    int status = R_ST_NORMAL;
 	    Str tok = Strnew();
-	    int status;
+	    int pre_mode = 0;
+	    int end_tag = 0;
 
 	    frame = f->frame[i];
 
@@ -536,20 +526,15 @@ createFrameFile(struct frameset *f, FILE * f1, Buffer *current, int level,
 		t_target = "_blank";
 		d_target = TargetSelf ? s_target : t_target;
 #ifdef JP_CHARSET
-		code = '\0';
-#endif				/* JP_CHARSET */
+		code = DocumentCode;
+#endif
 		t_stack = 0;
 		if (frame.body->type &&
 		    !strcasecmp(frame.body->type, "text/plain")) {
 		    Str tmp;
 		    fprintf(f1, "<pre>\n");
 		    while ((tmp = StrmyUFgets(&f2))->length) {
-#ifdef JP_CHARSET
-			if ((ic = checkShiftCode(tmp, code)) != '\0')
-			    tmp = conv_str(tmp, (code = ic), InnerCode);
-
-#endif				/* JP_CHARSET */
-			cleanup_line(tmp, HTML_MODE);
+			tmp = convertLine(NULL, tmp, &code, HTML_MODE);
 			fprintf(f1, "%s", html_quote(tmp->ptr));
 		    }
 		    fprintf(f1, "</pre>\n");
@@ -557,35 +542,88 @@ createFrameFile(struct frameset *f, FILE * f1, Buffer *current, int level,
 		    break;
 		}
 		do {
-		    status = R_ST_NORMAL;
+		    int is_tag = FALSE;
+		    char *q;
+		    struct parsed_tag *tag;
+
 		    do {
 			if (*p == '\0') {
 			    Str tmp = StrmyUFgets(&f2);
-			    if (tmp->length == 0 && status != R_ST_NORMAL)
-				tmp = correct_irrtag(status);
 			    if (tmp->length == 0)
 				break;
-#ifdef JP_CHARSET
-			    if ((ic = checkShiftCode(tmp, code)) != '\0')
-				tmp = conv_str(tmp, (code = ic), InnerCode);
-
-#endif				/* JP_CHARSET */
-			    cleanup_line(tmp, HTML_MODE);
+			    tmp = convertLine(NULL, tmp, &code, HTML_MODE);
 			    p = tmp->ptr;
 			}
-			if (status == R_ST_NORMAL || ST_IS_COMMENT(status))
-			    read_token(tok, &p, &status, 1, 0);
-			else
-			    read_token(tok, &p, &status, 1, 1);
+			read_token(tok, &p, &status, 1, status != R_ST_NORMAL);
 		    } while (status != R_ST_NORMAL);
 
 		    if (tok->length == 0)
 			continue;
 
 		    if (tok->ptr[0] == '<') {
+			if (tok->ptr[1] &&
+			    REALLY_THE_BEGINNING_OF_A_TAG(tok->ptr))
+			    is_tag = TRUE;
+			else if (!(pre_mode & (RB_PLAIN | RB_INTXTA |
+					       RB_SCRIPT | RB_STYLE))) {
+			    p = Strnew_m_charp(tok->ptr + 1, p, NULL)->ptr;
+			    tok = Strnew_charp("&lt;");
+			}
+		    }
+		    if (is_tag) {
+			if (pre_mode & (RB_PLAIN | RB_INTXTA | RB_SCRIPT |
+					RB_STYLE)) {
+			    q = tok->ptr;
+			    if ((tag = parse_tag(&q, FALSE)) &&
+				tag->tagid == end_tag) {
+				if (pre_mode & RB_PLAIN) {
+				    fputs("</PRE_PLAIN>", f1);
+				    pre_mode = 0;
+				    end_tag = 0;
+				    goto token_end;
+				}
+				pre_mode = 0;
+				end_tag = 0;
+				goto proc_normal;
+			    }
+			    if (strncmp(tok->ptr, "<!--", 4) &&
+				(q = strchr(tok->ptr + 1, '<'))) {
+				tok = Strnew_charp_n(tok->ptr, q - tok->ptr);
+				p = Strnew_m_charp(q, p, NULL)->ptr;
+				status = R_ST_NORMAL;
+			    }
+			    is_tag = FALSE;
+			}
+			else if (pre_mode & RB_INSELECT) {
+			    q = tok->ptr;
+			    if ((tag = parse_tag(&q, FALSE))) {
+				if ((tag->tagid == end_tag) ||
+				    (tag->tagid == HTML_N_FORM)) {
+				    if (tag->tagid == HTML_N_FORM)
+					fputs("</SELECT>", f1);
+				    pre_mode = 0;
+				    end_tag = 0;
+				    goto proc_normal;
+				}
+				if (t_stack) {
+				    switch (tag->tagid) {
+				    case HTML_TABLE:
+				    case HTML_N_TABLE:
+				      CASE_TABLE_TAG:
+					fputs("</SELECT>", f1);
+					pre_mode = 0;
+					end_tag = 0;
+					goto proc_normal;
+				    }
+				}
+			    }
+			}
+		    }
+
+		  proc_normal:
+		    if (is_tag) {
 			char *q = tok->ptr;
 			int j, a_target = 0;
-			struct parsed_tag *tag;
 			ParsedURL url;
 
 			if (!(tag = parse_tag(&q, FALSE)))
@@ -599,8 +637,9 @@ createFrameFile(struct frameset *f, FILE * f1, Buffer *current, int level,
 			    fputs("-->", f1);
 			    goto token_end;
 			case HTML_BASE:
+			    /* "BASE" is prohibit tag */
 			    if (parsedtag_get_value(tag, ATTR_HREF, &q)) {
-				q = url_quote_conv(q, code);
+				q = url_quote_conv(remove_space(q), code);
 				parseURL(q, &base, NULL);
 			    }
 			    if (parsedtag_get_value(tag, ATTR_TARGET, &q)) {
@@ -611,12 +650,31 @@ createFrameFile(struct frameset *f, FILE * f1, Buffer *current, int level,
 				else
 				    d_target = url_quote_conv(q, code);
 			    }
-			    /* fall thru, "BASE" is prohibit tag */
+			    Strshrinkfirst(tok, 1);
+			    Strshrink(tok, 1);
+			    fprintf(f1, "<!-- %s -->", html_quote(tok->ptr));
+			    goto token_end;
+			case HTML_META:
+			    if (parsedtag_get_value(tag, ATTR_HTTP_EQUIV, &q)
+				&& !strcasecmp(q, "refresh")) {
+				if (parsedtag_get_value(tag, ATTR_CONTENT, &q)
+				    ) {
+				    Str s_tmp = NULL;
+				    int refresh_interval =
+					getMetaRefreshParam(q, &s_tmp);
+				    if (s_tmp) {
+					q = html_quote(s_tmp->ptr);
+					fprintf(f1,
+						"Refresh (%d sec) <a href=\"%s\">%s</a>\n",
+						refresh_interval, q, q);
+				    }
+				}
+			    }
+			    /* fall thru, "META" is prohibit tag */
 			case HTML_HEAD:
 			case HTML_N_HEAD:
 			case HTML_BODY:
 			case HTML_N_BODY:
-			case HTML_META:
 			case HTML_DOCTYPE:
 			    /* prohibit_tags */
 			    Strshrinkfirst(tok, 1);
@@ -638,18 +696,7 @@ createFrameFile(struct frameset *f, FILE * f1, Buffer *current, int level,
 				goto token_end;
 			    }
 			    break;
-			case HTML_THEAD:
-			case HTML_N_THEAD:
-			case HTML_TBODY:
-			case HTML_N_TBODY:
-			case HTML_TFOOT:
-			case HTML_N_TFOOT:
-			case HTML_TD:
-			case HTML_N_TD:
-			case HTML_TR:
-			case HTML_N_TR:
-			case HTML_TH:
-			case HTML_N_TH:
+			  CASE_TABLE_TAG:
 			    /* table_tags MUST be in table stack */
 			    if (!t_stack) {
 				Strshrinkfirst(tok, 1);
@@ -660,6 +707,37 @@ createFrameFile(struct frameset *f, FILE * f1, Buffer *current, int level,
 
 			    }
 			    break;
+			case HTML_SELECT:
+			    pre_mode = RB_INSELECT;
+			    end_tag = HTML_N_SELECT;
+			    break;
+			case HTML_TEXTAREA:
+			    pre_mode = RB_INTXTA;
+			    end_tag = HTML_N_TEXTAREA;
+			    break;
+			case HTML_SCRIPT:
+			    pre_mode = RB_SCRIPT;
+			    end_tag = HTML_N_SCRIPT;
+			    break;
+			case HTML_STYLE:
+			    pre_mode = RB_STYLE;
+			    end_tag = HTML_N_STYLE;
+			    break;
+			case HTML_LISTING:
+			    pre_mode = RB_PLAIN;
+			    end_tag = HTML_N_LISTING;
+			    fputs("<PRE_PLAIN>", f1);
+			    goto token_end;
+			case HTML_XMP:
+			    pre_mode = RB_PLAIN;
+			    end_tag = HTML_N_XMP;
+			    fputs("<PRE_PLAIN>", f1);
+			    goto token_end;
+			case HTML_PLAINTEXT:
+			    pre_mode = RB_PLAIN;
+			    end_tag = MAX_HTMLTAG;
+			    fputs("<PRE_PLAIN>", f1);
+			    goto token_end;
 			default:
 			    break;
 			}
@@ -671,7 +749,9 @@ createFrameFile(struct frameset *f, FILE * f1, Buffer *current, int level,
 				if (!tag->value[j])
 				    break;
 				tag->value[j] =
-				    url_quote_conv(tag->value[j], code);
+				    url_quote_conv(remove_space(tag->value[j]),
+						   code);
+				tag->need_reconstruct = TRUE;
 				parseURL2(tag->value[j], &url, &base);
 				if (url.scheme == SCM_UNKNOWN ||
 #ifndef USE_W3MMAILER
@@ -681,7 +761,6 @@ createFrameFile(struct frameset *f, FILE * f1, Buffer *current, int level,
 				    break;
 				a_target |= 1;
 				tag->value[j] = parsedURL2Str(&url)->ptr;
-				tag->need_reconstruct = TRUE;
 				parsedtag_set_value(tag,
 						    ATTR_REFERER,
 						    parsedURL2Str(&base)->ptr);
@@ -721,16 +800,32 @@ createFrameFile(struct frameset *f, FILE * f1, Buffer *current, int level,
 			    parsedtag_set_value(tag, ATTR_TARGET, d_target);
 			}
 			if (parsedtag_need_reconstruct(tag))
-			    Strfputs(parsedtag2str(tag), f1);
-			else
-			    Strfputs(tok, f1);
+			    tok = parsedtag2str(tag);
+			Strfputs(tok, f1);
 		    }
 		    else {
-			Strfputs(tok, f1);
+			if (pre_mode & (RB_PLAIN | RB_INTXTA))
+			    fprintf(f1, "%s", html_quote(tok->ptr));
+			else
+			    Strfputs(tok, f1);
 		    }
 		  token_end:
 		    Strclear(tok);
 		} while (*p != '\0' || !iseos(f2.stream));
+		if (pre_mode & RB_PLAIN)
+		    fputs("</PRE_PLAIN>\n", f1);
+		else if (pre_mode & RB_INTXTA)
+		    fputs("</TEXTAREA></FORM>\n", f1);
+		else if (pre_mode & RB_INSELECT)
+		    fputs("</SELECT></FORM>\n", f1);
+		else if (pre_mode & (RB_SCRIPT | RB_STYLE)) {
+		    if (status != R_ST_NORMAL)
+			fputs(correct_irrtag(status)->ptr, f1);
+		    if (pre_mode & RB_SCRIPT)
+			fputs("</SCRIPT>\n", f1);
+		    else if (pre_mode & RB_STYLE)
+			fputs("</STYLE>\n", f1);
+		}
 		while (t_stack--)
 		    fputs("</TABLE>\n", f1);
 		UFclose(&f2);
@@ -752,9 +847,7 @@ createFrameFile(struct frameset *f, FILE * f1, Buffer *current, int level,
     fputs("</table>\n", f1);
     if (level == 0) {
 	fputs("</body></html>\n", f1);
-	signal(SIGINT, prevtrap);
-	if (fmInitialized)
-	    term_raw();
+	TRAP_OFF;
     }
     return 0;
 }
@@ -792,6 +885,7 @@ renderFrame(Buffer *Cbuf, int force_reload)
     buf->document_code = Cbuf->document_code;
 #endif
     copyParsedURL(&buf->currentURL, &Cbuf->currentURL);
+    preFormUpdateBuffer(buf);
     return buf;
 }
 
