@@ -1,4 +1,4 @@
-/* $Id: html.c,v 1.14 2002/10/16 18:11:15 ukai Exp $ */
+/* $Id: html.c,v 1.22 2002/12/09 15:40:36 ukai Exp $ */
 #include "html.h"
 
 /* Define HTML Tag Infomation Table */
@@ -7,9 +7,9 @@
 #define MAXA_CORE	1
 unsigned char ALST_A[] = {
     ATTR_NAME, ATTR_HREF, ATTR_TARGET, ATTR_HSEQ, ATTR_REFERER,
-    ATTR_FRAMENAME, ATTR_CORE
+    ATTR_FRAMENAME, ATTR_TITLE, ATTR_ACCESSKEY, ATTR_CORE
 };
-#define MAXA_A		MAXA_CORE + 6
+#define MAXA_A		MAXA_CORE + 8
 unsigned char ALST_P[] = { ATTR_ALIGN, ATTR_CORE };
 #define MAXA_P		MAXA_CORE + 1
 unsigned char ALST_UL[] = { ATTR_START, ATTR_TYPE, ATTR_CORE };
@@ -18,15 +18,19 @@ unsigned char ALST_LI[] = { ATTR_TYPE, ATTR_VALUE, ATTR_CORE };
 #define MAXA_LI		MAXA_CORE + 2
 unsigned char ALST_HR[] = { ATTR_WIDTH, ATTR_ALIGN, ATTR_CORE };
 #define MAXA_HR		MAXA_CORE + 2
+unsigned char ALST_LINK[] = { ATTR_HREF, ATTR_HSEQ, ATTR_REL, ATTR_REV,
+    ATTR_TITLE, ATTR_TYPE, ATTR_CORE
+};
+#define MAXA_LINK	MAXA_CORE + sizeof ALST_LINK/sizeof ALST_LINK[0] - 1
 unsigned char ALST_DL[] = { ATTR_COMPACT, ATTR_CORE };
 #define MAXA_DL		MAXA_CORE + 1
 unsigned char ALST_PRE[] = { ATTR_FOR_TABLE, ATTR_CORE };
 #define MAXA_PRE	MAXA_CORE + 1
 unsigned char ALST_IMG[] =
     { ATTR_SRC, ATTR_ALT, ATTR_WIDTH, ATTR_HEIGHT, ATTR_ALIGN, ATTR_USEMAP,
-    ATTR_ISMAP, ATTR_CORE
+    ATTR_ISMAP, ATTR_TITLE, ATTR_CORE
 };
-#define MAXA_IMG	MAXA_CORE + 7
+#define MAXA_IMG	MAXA_CORE + 8
 unsigned char ALST_TABLE[] =
     { ATTR_BORDER, ATTR_WIDTH, ATTR_HBORDER, ATTR_CELLSPACING,
     ATTR_CELLPADDING, ATTR_VSPACE, ATTR_CORE
@@ -64,8 +68,8 @@ unsigned char ALST_ISINDEX[] = { ATTR_ACTION, ATTR_PROMPT, ATTR_CORE };
 unsigned char ALST_MAP[] = { ATTR_NAME, ATTR_CORE };
 #define MAXA_MAP	MAXA_CORE + 1
 unsigned char ALST_AREA[] =
-    { ATTR_HREF, ATTR_ALT, ATTR_SHAPE, ATTR_COORDS, ATTR_CORE };
-#define MAXA_AREA	MAXA_CORE + 4
+    { ATTR_HREF, ATTR_TARGET, ATTR_ALT, ATTR_SHAPE, ATTR_COORDS, ATTR_CORE };
+#define MAXA_AREA	MAXA_CORE + 5
 unsigned char ALST_BASE[] = { ATTR_HREF, ATTR_TARGET, ATTR_CORE };
 #define MAXA_BASE	MAXA_CORE + 2
 unsigned char ALST_BODY[] = { ATTR_BACKGROUND, ATTR_CORE };
@@ -108,9 +112,10 @@ unsigned char ALST_INPUT_ALT[] =
 #define MAXA_INPUT_ALT  16
 unsigned char ALST_IMG_ALT[] =
     { ATTR_SRC, ATTR_WIDTH, ATTR_HEIGHT, ATTR_USEMAP, ATTR_ISMAP, ATTR_HSEQ,
-    ATTR_XOFFSET, ATTR_YOFFSET, ATTR_TOP_MARGIN, ATTR_BOTTOM_MARGIN
+    ATTR_XOFFSET, ATTR_YOFFSET, ATTR_TOP_MARGIN, ATTR_BOTTOM_MARGIN,
+    ATTR_TITLE
 };
-#define MAXA_IMG_ALT  10
+#define MAXA_IMG_ALT  11
 unsigned char ALST_NOP[] = { ATTR_CORE };
 #define MAXA_NOP	MAXA_CORE
 
@@ -220,12 +225,12 @@ TagInfo TagMAP[MAX_HTMLTAG] = {
     {"/sup", NULL, 0, 0},	/* 101 HTML_N_SUP       */
     {"sub", NULL, 0, 0},	/* 102 HTML_SUB       */
     {"/sub", NULL, 0, 0},	/* 103 HTML_N_SUB       */
-    {NULL, NULL, 0, 0},		/* 104 Undefined       */
-    {NULL, NULL, 0, 0},		/* 105 Undefined       */
+    {"link", ALST_LINK, MAXA_LINK, 0},	/*  104 HTML_LINK      */
+    {"s", NULL, 0, 0},		/*  105 HTML_S        */
+    {"/s", NULL, 0, TFLG_END},	/*  106 HTML_N_S      */
+    {NULL, NULL, 0, 0},		/* 107 Undefined       */
 
     /* pseudo tag */
-    {"internal", NULL, 0, TFLG_INT},	/* 106 HTML_INTERNAL   */
-    {"/internal", NULL, 0, TFLG_INT | TFLG_END},	/* 107 HTML_N_INTERNAL   */
     {"select_int", ALST_SELECT_INT, MAXA_SELECT_INT, TFLG_INT},	/* 108 HTML_SELECT_INT   */
     {"/select_int", NULL, 0, TFLG_INT | TFLG_END},	/* 109 HTML_N_SELECT_INT */
     {"option_int", ALST_OPTION, MAXA_OPTION, TFLG_INT},	/* 110 HTML_OPTION_INT   */
@@ -244,8 +249,13 @@ TagInfo TagMAP[MAX_HTMLTAG] = {
     {"/input_alt", NULL, 0, TFLG_INT | TFLG_END},	/* 123 HTML_N_INPUT_ALT */
     {"img_alt", ALST_IMG_ALT, MAXA_IMG_ALT, TFLG_INT},	/* 124 HTML_IMG_ALT     */
     {"/img_alt", NULL, 0, TFLG_INT | TFLG_END},	/* 125 HTML_N_IMG_ALT   */
-    {"eol", NULL, 0, TFLG_INT},	/* 126 HTML_EOL         */
-    {" ", ALST_NOP, MAXA_NOP, TFLG_INT},	/* 127 HTML_NOP         */
+    {" ", ALST_NOP, MAXA_NOP, TFLG_INT},	/* 126 HTML_NOP         */
+    {"pre_plain", NULL, 0, TFLG_INT},	/* 127 HTML_PRE_PLAIN         */
+    {"/pre_plain", NULL, 0, TFLG_INT | TFLG_END},	/* 128 HTML_N_PRE_PLAIN         */
+    {"internal", NULL, 0, TFLG_INT},	/* 129 HTML_INTERNAL   */
+    {"/internal", NULL, 0, TFLG_INT | TFLG_END},	/* 130 HTML_N_INTERNAL   */
+    {"div_int", ALST_P, MAXA_P, TFLG_INT},	/*  131 HTML_DIV_INT    */
+    {"/div_int", NULL, 0, TFLG_INT | TFLG_END},	/*  132 HTML_N_DIV_INT  */
 };
 
 TagAttrInfo AttrMAP[MAX_TAGATTR] = {
@@ -296,24 +306,33 @@ TagAttrInfo AttrMAP[MAX_TAGATTR] = {
     {"shape", VTYPE_STR, 0},	/* 44 ATTR_SHAPE          */
     {"coords", VTYPE_STR, 0},	/* 45 ATTR_COORDS         */
     {"ismap", VTYPE_NONE, 0},	/* 46 ATTR_ISMAP          */
-
-    {NULL, VTYPE_NONE, 0},	/* 47 Undefined           */
-    {NULL, VTYPE_NONE, 0},	/* 48 Undefined           */
+    {"rel", VTYPE_STR, 0},	/* 47 ATTR_REL            */
+    {"rev", VTYPE_STR, 0},	/* 48 ATTR_REV            */
+    {"title", VTYPE_STR, 0},	/* 49 ATTR_TITLE          */
+    {"accesskey", VTYPE_STR, 0},	/* 50 ATTR_ACCESSKEY          */
+    {NULL, VTYPE_NONE, 0},	/* 51 Undefined           */
+    {NULL, VTYPE_NONE, 0},	/* 52 Undefined           */
+    {NULL, VTYPE_NONE, 0},	/* 53 Undefined           */
+    {NULL, VTYPE_NONE, 0},	/* 54 Undefined           */
+    {NULL, VTYPE_NONE, 0},	/* 55 Undefined           */
+    {NULL, VTYPE_NONE, 0},	/* 56 Undefined           */
+    {NULL, VTYPE_NONE, 0},	/* 57 Undefined           */
+    {NULL, VTYPE_NONE, 0},	/* 58 Undefined           */
+    {NULL, VTYPE_NONE, 0},	/* 59 Undefined           */
 
     /* Internal attribute */
-    {"xoffset", VTYPE_NUMBER, AFLG_INT},	/* 49 ATTR_XOFFSET        */
-    {"yoffset", VTYPE_NUMBER, AFLG_INT},	/* 50 ATTR_YOFFSET        */
-    {"top_margin", VTYPE_NUMBER, AFLG_INT},	/* 51 ATTR_TOP_MARGIN,    */
-    {"bottom_margin", VTYPE_NUMBER, AFLG_INT},	/* 52 ATTR_BOTTOM_MARGIN, */
-    {"tid", VTYPE_NUMBER, AFLG_INT},	/* 53 ATTR_TID            */
-    {"fid", VTYPE_NUMBER, AFLG_INT},	/* 54 ATTR_FID            */
-    {"for_table", VTYPE_NONE, AFLG_INT},	/* 55 ATTR_FOR_TABLE      */
-    {"framename", VTYPE_STR, AFLG_INT},	/* 56 ATTR_FRAMENAME      */
-    {"hborder", VTYPE_NONE, 0},	/* 57 ATTR_HBORDER        */
-    {"hseq", VTYPE_NUMBER, AFLG_INT},	/* 58 ATTR_HSEQ           */
-    {"no_effect", VTYPE_NONE, AFLG_INT},	/* 59 ATTR_NO_EFFECT      */
-    {"referer", VTYPE_STR, AFLG_INT},	/* 60 ATTR_REFERER        */
-    {"selectnumber", VTYPE_NUMBER, AFLG_INT},	/* 61 ATTR_SELECTNUMBER   */
-    {"textareanumber", VTYPE_NUMBER, AFLG_INT},	/* 62 ATTR_TEXTAREANUMBER */
-    {"title", VTYPE_STR, AFLG_INT},	/* 63 ATTR_TITLE          */
+    {"xoffset", VTYPE_NUMBER, AFLG_INT},	/* 60 ATTR_XOFFSET        */
+    {"yoffset", VTYPE_NUMBER, AFLG_INT},	/* 61 ATTR_YOFFSET        */
+    {"top_margin", VTYPE_NUMBER, AFLG_INT},	/* 62 ATTR_TOP_MARGIN,    */
+    {"bottom_margin", VTYPE_NUMBER, AFLG_INT},	/* 63 ATTR_BOTTOM_MARGIN, */
+    {"tid", VTYPE_NUMBER, AFLG_INT},	/* 64 ATTR_TID            */
+    {"fid", VTYPE_NUMBER, AFLG_INT},	/* 65 ATTR_FID            */
+    {"for_table", VTYPE_NONE, AFLG_INT},	/* 66 ATTR_FOR_TABLE      */
+    {"framename", VTYPE_STR, AFLG_INT},	/* 67 ATTR_FRAMENAME      */
+    {"hborder", VTYPE_NONE, 0},	/* 68 ATTR_HBORDER        */
+    {"hseq", VTYPE_NUMBER, AFLG_INT},	/* 69 ATTR_HSEQ           */
+    {"no_effect", VTYPE_NONE, AFLG_INT},	/* 70 ATTR_NO_EFFECT      */
+    {"referer", VTYPE_STR, AFLG_INT},	/* 71 ATTR_REFERER        */
+    {"selectnumber", VTYPE_NUMBER, AFLG_INT},	/* 72 ATTR_SELECTNUMBER   */
+    {"textareanumber", VTYPE_NUMBER, AFLG_INT},	/* 73 ATTR_TEXTAREANUMBER */
 };
