@@ -1,4 +1,4 @@
-/* $Id: indep.h,v 1.9 2002/08/20 17:49:39 ukai Exp $ */
+/* $Id: indep.h,v 1.13 2003/02/05 16:43:57 ukai Exp $ */
 #ifndef INDEP_H
 #define INDEP_H
 #include "gc.h"
@@ -12,9 +12,25 @@
 #define FALSE 0
 #endif				/* FALSE */
 
-#define PAGER_MODE	0
-#define HTML_MODE	1
-#define HEADER_MODE	2
+#define RAW_MODE	0
+#define PAGER_MODE	1
+#define HTML_MODE	2
+#define HEADER_MODE	3
+
+extern unsigned char QUOTE_MAP[];
+extern char *HTML_QUOTE_MAP[];
+#define HTML_QUOTE_MASK   0x07	/* &, <, >, " */
+#define SHELL_UNSAFE_MASK 0x08	/* [^A-Za-z0-9_./:\200-\377] */
+#define URL_QUOTE_MASK    0x10	/* [\0- \177-\377] */
+#define FILE_QUOTE_MASK   0x30	/* [\0- #%&+:?\177-\377] */
+#define URL_UNSAFE_MASK   0x70	/* [^A-Za-z0-9_$\-.] */
+#define GET_QUOTE_TYPE(c) QUOTE_MAP[(int)(unsigned char)(c)]
+#define is_html_quote(c)   (GET_QUOTE_TYPE(c) & HTML_QUOTE_MASK)
+#define is_shell_unsafe(c) (GET_QUOTE_TYPE(c) & SHELL_UNSAFE_MASK)
+#define is_url_quote(c)    (GET_QUOTE_TYPE(c) & URL_QUOTE_MASK)
+#define is_file_quote(c)   (GET_QUOTE_TYPE(c) & FILE_QUOTE_MASK)
+#define is_url_unsafe(c)   (GET_QUOTE_TYPE(c) & URL_UNSAFE_MASK)
+#define html_quote_char(c) HTML_QUOTE_MAP[(int)is_html_quote(c)]
 
 extern clen_t strtoclen(const char *s);
 extern char *conv_entity(int ch);
@@ -39,17 +55,17 @@ extern int strcasemstr(char *str, char *srch[], char **ret_ptr);
 extern char *remove_space(char *str);
 extern int non_null(char *s);
 extern void cleanup_line(Str s, int mode);
-extern char *html_quote_char(char c);
 extern char *html_quote(char *str);
 extern char *html_unquote(char *str);
 extern char *file_quote(char *str);
 extern char *file_unquote(char *str);
 extern char *url_quote(char *str);
-extern char *url_unquote(char *str);
+extern Str Str_url_unquote(Str x, int is_form, int safe);
 extern Str Str_form_quote(Str x);
-extern Str Str_form_unquote(Str x);
+#define Str_form_unquote(x) Str_url_unquote((x), TRUE, FALSE)
 extern char *shell_quote(char *str);
 
+extern char *w3m_auxbin_dir();
 extern char *w3m_lib_dir();
 extern char *w3m_etc_dir();
 extern char *w3m_help_dir();
