@@ -1,9 +1,10 @@
-/* $Id: local.c,v 1.13 2001/12/03 18:29:37 ukai Exp $ */
+/* $Id: local.c,v 1.28 2003/02/08 18:23:53 ukai Exp $ */
 #include "fm.h"
 #include <string.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <sys/types.h>
+#include <sys/stat.h>
 #include <signal.h>
 #include <errno.h>
 #ifdef HAVE_READLINK
@@ -13,28 +14,45 @@
 #include <limits.h>		/* _MAX_PATH ? */
 #endif				/* __EMX__ */
 #include "local.h"
+#include "hash.h"
 
 #define CGIFN_NORMAL     0
-#define CGIFN_DROOT      1
+#define CGIFN_LIBDIR     1
 #define CGIFN_CGIBIN     2
-#define CGIFN_MODE(x) ((x)&3)
 
-#define CGIFN_CONTAIN_SLASH 4
+static char *Local_cookie_file = NULL;
+
+static void
+writeLocalCookie()
+{
+    FILE *f;
+
+    if (no_rc_dir)
+	return;
+    if (Local_cookie_file)
+	return;
+    Local_cookie_file = tmpfname(TMPF_COOKIE, NULL)->ptr;
+    set_environ("LOCAL_COOKIE_FILE", Local_cookie_file);
+    f = fopen(Local_cookie_file, "wb");
+    if (!f)
+	return;
+    fwrite(Local_cookie->ptr, sizeof(char), Local_cookie->length, f);
+    fclose(f);
+    chmod(Local_cookie_file, S_IRUSR | S_IWUSR);
+}
 
 /* setup cookie for local CGI */
 void
 setLocalCookie()
 {
-    Str buf;
     char hostname[256];
     gethostname(hostname, 256);
 
-    buf = Sprintf("%d.%ld@%s", getpid(), lrand48(), hostname);
-    Local_cookie = buf->ptr;
+    Local_cookie = Sprintf("%d.%ld@%s", CurrentPid, lrand48(), hostname);
 }
 
-Buffer *
-dirBuffer(char *dname)
+Str
+loadLocalDir(char *dname)
 {
     Str tmp;
     DIR *d;
@@ -43,24 +61,27 @@ dirBuffer(char *dname)
     char **flist;
     char *p, *qdir;
     Str fbuf = Strnew();
-#ifdef HAVE_READLINK
+#ifdef HAVE_LSTAT
     struct stat lst;
+#ifdef HAVE_READLINK
     char lbuf[1024];
 #endif				/* HAVE_READLINK */
+#endif				/* HAVE_LSTAT */
     int i, l, nrow = 0, n = 0, maxlen = 0;
     int nfile, nfile_max = 100;
     Str dirname;
-    Buffer *buf;
 
     d = opendir(dname);
     if (d == NULL)
 	return NULL;
     dirname = Strnew_charp(dname);
+    if (Strlastchar(dirname) != '/')
+	Strcat_char(dirname, '/');
     qdir = html_quote(Str_conv_from_system(dirname)->ptr);
-    tmp =
-	Sprintf
-	("<title>Directory list of %s</title><h1>Directory list of %s</h1>\n",
-	 qdir, qdir);
+    tmp = Strnew_m_charp("<HTML>\n<HEAD>\n<BASE HREF=\"file://", qdir,
+			 "\">\n<TITLE>Directory list of ", qdir,
+			 "</TITLE>\n</HEAD>\n<BODY>\n<H1>Directory list of ",
+			 qdir, "</H1>\n", NULL);
     flist = New_N(char *, nfile_max);
     nfile = 0;
     while ((dir = readdir(d)) != NULL) {
@@ -83,7 +104,7 @@ dirBuffer(char *dname)
 	    l = 1;
 	nrow = (n + l - 1) / l;
 	n = 1;
-	Strcat_charp(tmp, "<TABLE CELLPADDING=0><TR VALIGN=TOP>\n");
+	Strcat_charp(tmp, "<TABLE CELLPADDING=0>\n<TR VALIGN=TOP>\n");
     }
     qsort((void *)flist, nfile, sizeof(char *), strCmp);
     for (i = 0; i < nfile; i++) {
@@ -94,10 +115,10 @@ dirBuffer(char *dname)
 	if (Strlastchar(fbuf) != '/')
 	    Strcat_char(fbuf, '/');
 	Strcat_charp(fbuf, p);
-#ifdef HAVE_READLINK		/* readlink == lstat() ? (ukai) */
+#ifdef HAVE_LSTAT
 	if (lstat(fbuf->ptr, &lst) < 0)
 	    continue;
-#endif				/* READLINK */
+#endif				/* HAVE_LSTAT */
 	if (stat(fbuf->ptr, &st) < 0)
 	    continue;
 	if (multicolList) {
@@ -105,22 +126,23 @@ dirBuffer(char *dname)
 		Strcat_charp(tmp, "<TD><NOBR>");
 	}
 	else {
+#ifdef HAVE_LSTAT
+	    if (S_ISLNK(lst.st_mode))
+		Strcat_charp(tmp, "[LINK] ");
+	    else
+#endif				/* HAVE_LSTAT */
 	    if (S_ISDIR(st.st_mode))
 		Strcat_charp(tmp, "[DIR]&nbsp; ");
-#ifdef HAVE_READLINK
-	    else if (S_ISLNK(lst.st_mode))
-		Strcat_charp(tmp, "[LINK] ");
-#endif				/* HAVE_READLINE */
 	    else
 		Strcat_charp(tmp, "[FILE] ");
 	}
-	Strcat_m_charp(tmp, "<A HREF=\"", file_to_url(fbuf->ptr), NULL);
+	Strcat_m_charp(tmp, "<A HREF=\"", html_quote(file_quote(p)), NULL);
 	if (S_ISDIR(st.st_mode))
 	    Strcat_char(tmp, '/');
 	Strcat_m_charp(tmp, "\">", html_quote(conv_from_system(p)), NULL);
 	if (S_ISDIR(st.st_mode))
 	    Strcat_char(tmp, '/');
-	Strcat_charp(tmp, "</a>");
+	Strcat_charp(tmp, "</A>");
 	if (multicolList) {
 	    if (n++ == nrow) {
 		Strcat_charp(tmp, "</NOBR></TD>\n");
@@ -131,7 +153,7 @@ dirBuffer(char *dname)
 	    }
 	}
 	else {
-#ifdef HAVE_READLINK
+#if defined(HAVE_LSTAT) && defined(HAVE_READLINK)
 	    if (S_ISLNK(lst.st_mode)) {
 		if ((l = readlink(fbuf->ptr, lbuf, sizeof(lbuf))) > 0) {
 		    lbuf[l] = '\0';
@@ -141,20 +163,16 @@ dirBuffer(char *dname)
 			Strcat_char(tmp, '/');
 		}
 	    }
-#endif				/* HAVE_READLINK */
+#endif				/* HAVE_LSTAT && HAVE_READLINK */
 	    Strcat_charp(tmp, "<br>\n");
 	}
     }
     if (multicolList) {
-	Strcat_charp(tmp, "</TR></TABLE>\n");
+	Strcat_charp(tmp, "</TR>\n</TABLE>\n");
     }
+    Strcat_charp(tmp, "</BODY>\n</HTML>\n");
 
-    buf = loadHTMLString(tmp);
-#ifdef JP_CHARSET
-    if (buf)
-	buf->document_code = SystemCode;
-#endif
-    return buf;
+    return tmp;
 }
 
 static int
@@ -162,45 +180,14 @@ check_local_cgi(char *file, int status)
 {
     struct stat st;
 
-    if (status & CGIFN_CONTAIN_SLASH) {
-	/* local CGI file must be just under /cgi-bin/
-	 * or /$LIB/
-	 */
+    if (status != CGIFN_LIBDIR && status != CGIFN_CGIBIN)
 	return -1;
-    }
-#ifdef __EMX__
-    if (CGIFN_MODE(status) != CGIFN_CGIBIN) {
-	char tmp[_MAX_PATH];
-	int len;
-
-	_abspath(tmp, w3m_lib_dir(), _MAX_PATH);	/* Translate '\\' to '/' */
-	len = strlen(tmp);
-	while (len > 1 && tmp[len - 1] == '/')
-	    len--;
-	if (strnicmp(file, tmp, len) ||	/* and ignore case  */
-	    (file[len] != '/'))
-	    return -1;
-    }
-#else				/* not __EMX__ */
-    if (CGIFN_MODE(status) != CGIFN_CGIBIN) {
-	char *tmp = Strnew_charp(w3m_lib_dir())->ptr;
-	int len = strlen(tmp);
-
-	while (len > 1 && tmp[len - 1] == '/')
-	    len--;
-	if (strncmp(file, tmp, len) || (file[len] != '/'))
-	    /* 
-	     * a local-CGI script should be located on either
-	     * /cgi-bin/ directory or LIB_DIR (typically /usr/local/lib/w3m).
-	     */
-	    return -1;
-    }
-#endif				/* not __EMX__ */
     if (stat(file, &st) < 0)
 	return -1;
-    if ((st.st_uid == geteuid() && (st.st_mode & S_IXUSR)) || (st.st_gid == getegid() && (st.st_mode & S_IXGRP)) || (st.st_mode & S_IXOTH)) {	/* executable */
+    if (S_ISDIR(st.st_mode))
+	return -1;
+    if ((st.st_uid == geteuid() && (st.st_mode & S_IXUSR)) || (st.st_gid == getegid() && (st.st_mode & S_IXGRP)) || (st.st_mode & S_IXOTH))	/* executable */
 	return 0;
-    }
     return -1;
 }
 
@@ -212,7 +199,12 @@ set_environ(char *var, char *value)
 	setenv(var, value, 1);
 #else				/* not HAVE_SETENV */
 #ifdef HAVE_PUTENV
+    static Hash_sv *env_hash = NULL;
     Str tmp = Strnew_m_charp(var, "=", value, NULL);
+
+    if (env_hash == NULL)
+	env_hash = newHash_sv(20);
+    putHash_sv(env_hash, var, (void *)tmp->ptr);
     putenv(tmp->ptr);
 #else				/* not HAVE_PUTENV */
     extern char **environ;
@@ -265,174 +257,146 @@ set_cgi_environ(char *name, char *fn, char *req_uri)
     set_environ("SCRIPT_NAME", name);
     set_environ("SCRIPT_FILENAME", fn);
     set_environ("REQUEST_URI", req_uri);
-    set_environ("LOCAL_COOKIE", Local_cookie);
 }
 
 static Str
 checkPath(char *fn, char *path)
 {
+    char *p;
     Str tmp;
     struct stat st;
     while (*path) {
-	tmp = Strnew();
-	while (*path && *path != ':')
-	    Strcat_char(tmp, *path++);
-	if (*path == ':')
-	    path++;
+	p = strchr(path, ':');
+	tmp = Strnew_charp(expandPath(p ? allocStr(path, p - path) : path));
 	if (Strlastchar(tmp) != '/')
 	    Strcat_char(tmp, '/');
 	Strcat_charp(tmp, fn);
 	if (stat(tmp->ptr, &st) == 0)
 	    return tmp;
+	if (!p)
+	    break;
+	path = p + 1;
+	while (*path == ':')
+	    path++;
     }
     return NULL;
 }
 
-static char *
-cgi_filename(char *fn, int *status)
+static int
+cgi_filename(char *uri, char **fn, char **name, char **path_info)
 {
     Str tmp;
-    struct stat st;
-    if (cgi_bin != NULL && strncmp(fn, "/cgi-bin/", 9) == 0) {
-	*status = CGIFN_CGIBIN;
-	if (strchr(fn + 9, '/'))
-	    *status |= CGIFN_CONTAIN_SLASH;
-	tmp = checkPath(fn + 9, cgi_bin);
+    int offset;
+
+    *fn = uri;
+    *name = uri;
+    *path_info = NULL;
+
+    if (cgi_bin != NULL && strncmp(uri, "/cgi-bin/", 9) == 0) {
+	offset = 9;
+	if ((*path_info = strchr(uri + offset, '/')))
+	    *name = allocStr(uri, *path_info - uri);
+	tmp = checkPath(*name + offset, cgi_bin);
 	if (tmp == NULL)
-	    return fn;
-	return tmp->ptr;
-    }
-    if (strncmp(fn, "/$LIB/", 6) == 0) {
-	*status = CGIFN_NORMAL;
-	tmp = Strnew_charp(w3m_lib_dir());
-	fn += 5;
-	if (strchr(fn + 1, '/'))
-	    *status |= CGIFN_CONTAIN_SLASH;
-	if (Strlastchar(tmp) == '/')
-	    fn++;
-	Strcat_charp(tmp, fn);
-	return tmp->ptr;
-    }
-    if (*fn == '/' && document_root != NULL && stat(fn, &st) < 0) {
-	*status = CGIFN_DROOT;
-	if (strchr(fn + 1, '/'))
-	    *status |= CGIFN_CONTAIN_SLASH;
-	tmp = Strnew_charp(document_root);
-	if (Strlastchar(tmp) != '/')
-	    Strcat_char(tmp, '/');
-	Strcat_charp(tmp, fn);
-	return tmp->ptr;
-    }
-    *status = CGIFN_NORMAL;
-    return fn;
-}
-
-static pid_t
-localcgi_popen_r(FILE ** p_fp)
-{
-    int fd[2];
-    FILE *fp;
-    pid_t pid;
-    Str emsg;
-
-    if (pipe(fd) < 0) {
-	emsg = Sprintf("localcgi_popen_r: pipe: %s", strerror(errno));
-	disp_err_message(emsg->ptr, FALSE);
-	return (pid_t) - 1;
+	    return CGIFN_NORMAL;
+	*fn = tmp->ptr;
+	return CGIFN_CGIBIN;
     }
 
-    flush_tty();
-    if ((pid = fork()) < 0) {
-	emsg = Sprintf("localcgi_popen_r: fork: %s", strerror(errno));
-	disp_err_message(emsg->ptr, FALSE);
-	close(fd[0]);
-	close(fd[1]);
-	return (pid_t) - 1;
+#ifdef __EMX__
+    {
+	char lib[_MAX_PATH];
+	_abspath(lib, w3m_lib_dir(), _MAX_PATH);	/* Translate '\\' to '/' */
+	tmp = Strnew_charp(lib);
     }
-    else if (!pid) {
-	close_tty();
-	dup2(fd[1], 1);
-
-	if (fd[1] > 1)
-	    close(fd[1]);
-
-	close(fd[0]);
+#else
+    tmp = Strnew_charp(w3m_lib_dir());
+#endif
+    if (Strlastchar(tmp) != '/')
+	Strcat_char(tmp, '/');
+    if (strncmp(uri, "/$LIB/", 6) == 0)
+	offset = 6;
+    else if (strncmp(uri, tmp->ptr, tmp->length) == 0)
+	offset = tmp->length;
+    else if (*uri == '/' && document_root != NULL) {
+	Str tmp2 = Strnew_charp(document_root);
+	if (Strlastchar(tmp2) != '/')
+	    Strcat_char(tmp2, '/');
+	Strcat_charp(tmp2, uri + 1);
+	if (strncmp(tmp2->ptr, tmp->ptr, tmp->length) != 0)
+	    return CGIFN_NORMAL;
+	uri = tmp2->ptr;
+	*name = uri;
+	offset = tmp->length;
     }
-    else {
-	close(fd[1]);
-
-	if (!(fp = fdopen(fd[0], "r"))) {
-	    emsg =
-		Sprintf("localcgi_popen_r: fdopen(%d, \"r\"): %s", fd[0],
-			strerror(errno));
-	    disp_err_message(emsg->ptr, FALSE);
-	    kill(pid, SIGTERM);
-	    close(fd[0]);
-	    return (pid_t) - 1;
-	}
-
-	*p_fp = fp;
-    }
-
-    return pid;
+    else
+	return CGIFN_NORMAL;
+    if ((*path_info = strchr(uri + offset, '/')))
+	*name = allocStr(uri, *path_info - uri);
+    Strcat_charp(tmp, *name + offset);
+    *fn = tmp->ptr;
+    return CGIFN_LIBDIR;
 }
 
 FILE *
 localcgi_post(char *uri, char *qstr, FormList *request, char *referer)
 {
-    FILE *f, *f1;
-    Str tmp1;
+    FILE *fr = NULL, *fw = NULL;
     int status;
     pid_t pid;
-    char *file;
+    char *file = uri, *name = uri, *path_info = NULL, *tmpf = NULL;
 
-    file = cgi_filename(uri, &status);
+    status = cgi_filename(uri, &file, &name, &path_info);
     if (check_local_cgi(file, status) < 0)
 	return NULL;
-    tmp1 = tmpfname(TMPF_DFL, NULL);
-    f1 = fopen(tmp1->ptr, "w");
-    if (f1 == NULL)
+    writeLocalCookie();
+    if (request && request->enctype != FORM_ENCTYPE_MULTIPART) {
+	tmpf = tmpfname(TMPF_DFL, NULL)->ptr;
+	fw = fopen(tmpf, "w");
+	if (!fw)
+	    return NULL;
+    }
+    pid = open_pipe_rw(&fr, NULL);
+    if (pid < 0)
 	return NULL;
-    pushText(fileToDelete, tmp1->ptr);
-    if ((pid = localcgi_popen_r(&f))) {
-	fclose(f1);
-	return pid > 0 ? f : NULL;
+    else if (pid) {
+	if (fw)
+	    fclose(fw);
+	return fr;
     }
-    if (qstr == NULL) {
-	set_cgi_environ(uri, file, uri);
-    }
-    else {
-	set_cgi_environ(uri, file, Strnew_m_charp(uri, "?", qstr, NULL)->ptr);
-    }
-    set_environ("REQUEST_METHOD", "POST");
+    setup_child(TRUE, 2, fw ? fileno(fw) : -1);
+
     if (qstr)
-	set_environ("QUERY_STRING", qstr);
-    set_environ("CONTENT_LENGTH", Sprintf("%d", request->length)->ptr);
+	uri = Strnew_m_charp(uri, "?", qstr, NULL)->ptr;
+    set_cgi_environ(name, file, uri);
+    if (path_info)
+	set_environ("PATH_INFO", path_info);
     if (referer && referer != NO_REFERER)
 	set_environ("HTTP_REFERER", referer);
-    if (request->enctype == FORM_ENCTYPE_MULTIPART) {
-	set_environ("CONTENT_TYPE",
-		    Sprintf("multipart/form-data; boundary=%s",
-			    request->boundary)->ptr);
-    }
-    else {
-	set_environ("CONTENT_TYPE", "application/x-www-form-urlencoded");
-    }
-    if (request->enctype == FORM_ENCTYPE_MULTIPART) {
-	FILE *fd;
-	int c;
-	fd = fopen(request->body, "r");
-	if (fd != NULL) {
-	    while ((c = fgetc(fd)) != EOF)
-		fputc(c, f1);
-	    fclose(fd);
+    if (request) {
+	set_environ("REQUEST_METHOD", "POST");
+	if (qstr)
+	    set_environ("QUERY_STRING", qstr);
+	set_environ("CONTENT_LENGTH", Sprintf("%d", request->length)->ptr);
+	if (request->enctype == FORM_ENCTYPE_MULTIPART) {
+	    set_environ("CONTENT_TYPE",
+			Sprintf("multipart/form-data; boundary=%s",
+				request->boundary)->ptr);
+	    freopen(request->body, "r", stdin);
+	}
+	else {
+	    set_environ("CONTENT_TYPE", "application/x-www-form-urlencoded");
+	    fwrite(request->body, sizeof(char), request->length, fw);
+	    fclose(fw);
+	    freopen(tmpf, "r", stdin);
 	}
     }
     else {
-	fputs(request->body, f1);
+	set_environ("REQUEST_METHOD", "GET");
+	set_environ("QUERY_STRING", qstr ? qstr : "");
+	freopen(DEV_NULL_PATH, "r", stdin);
     }
-    fclose(f1);
-    freopen(tmp1->ptr, "r", stdin);
+
 #ifdef HAVE_CHDIR		/* ifndef __EMX__ ? */
     chdir(mydirname(file));
 #endif
@@ -442,40 +406,3 @@ localcgi_post(char *uri, char *qstr, FormList *request, char *referer)
     exit(1);
     return NULL;
 }
-
-FILE *
-localcgi_get(char *uri, char *request, char *referer)
-{
-    FILE *f;
-    int status;
-    pid_t pid;
-    char *file;
-
-    file = cgi_filename(uri, &status);
-    if (check_local_cgi(file, status) < 0)
-	return NULL;
-    if ((pid = localcgi_popen_r(&f)) < 0)
-	return NULL;
-    else if (pid)
-	return f;
-    if (request == NULL) {
-	set_cgi_environ(Strnew_charp(uri)->ptr, file, Strnew_charp(uri)->ptr);
-    }
-    else {
-	set_cgi_environ(Strnew_charp(uri)->ptr, file,
-			Strnew_m_charp(uri, "?", request, NULL)->ptr);
-    }
-    if (referer && referer != NO_REFERER)
-	set_environ("HTTP_REFERER", referer);
-    set_environ("REQUEST_METHOD", "GET");
-    set_environ("QUERY_STRING", request ? request : "");
-    freopen(DEV_NULL_PATH, "r", stdin);
-#ifdef HAVE_CHDIR		/* ifndef __EMX__? */
-    chdir(mydirname(file));
-#endif
-    execl(file, mybasename(file), NULL);
-    fprintf(stderr, "execl(\"%s\", \"%s\", NULL): %s\n",
-	    file, mybasename(file), strerror(errno));
-    exit(1);
-    return NULL;
-}
