@@ -1,11 +1,24 @@
-/* $Id: map.c,v 1.8 2002/03/15 18:48:55 ukai Exp $ */
+/* $Id: map.c,v 1.28 2003/02/05 16:45:08 ukai Exp $ */
 /*
  * client-side image maps
  */
 #include "fm.h"
 #include <math.h>
 
-#ifdef MENU_MAP
+MapList *
+searchMapList(Buffer *buf, char *name)
+{
+    MapList *ml;
+
+    if (name == NULL)
+	return NULL;
+    for (ml = buf->maplist; ml != NULL; ml = ml->next) {
+	if (!Strcmp_charp(ml->name, name))
+	    break;
+    }
+    return ml;
+}
+
 #ifdef USE_IMAGE
 static int
 inMapArea(MapArea * a, int x, int y)
@@ -57,123 +70,98 @@ inMapArea(MapArea * a, int x, int y)
     }
     return FALSE;
 }
-#endif
 
-char *
-follow_map_menu(Buffer *buf, struct parsed_tagarg *arg, Anchor *a_img, int x,
-		int y)
+static int
+nearestMapArea(MapList *ml, int x, int y)
 {
-    MapList *ml;
     ListItem *al;
     MapArea *a;
-    char *name;
-    int i, n, selected = -1, initial;
-    char **label;
-#ifdef USE_IMAGE
-    int px, py, map = 0;
-#endif
+    int i, l, n = -1, min = -1, limit = pixel_per_char * pixel_per_char
+	+ pixel_per_line * pixel_per_line;
 
-    name = tag_get_value(arg, "link");
-    if (name == NULL)
-	return NULL;
-
-    for (ml = buf->maplist; ml != NULL; ml = ml->next) {
-	if (!Strcmp_charp(ml->name, name))
-	    break;
-    }
-    if (ml == NULL || ml->area == NULL)
-	return NULL;
-
-    n = ml->area->nitem;
-    if (n == 0)
-	return NULL;
-    label = New_N(char *, n + 1);
-#ifdef USE_IMAGE
-    if (getMapXY(buf, a_img, &px, &py))
-	map = 1;
-#endif
-    initial = -n;
+    if (!ml || !ml->area)
+	return n;
     for (i = 0, al = ml->area->first; al != NULL; i++, al = al->next) {
 	a = (MapArea *) al->ptr;
 	if (a) {
-	    label[i] = *a->alt ? a->alt : a->url;
-#ifdef USE_IMAGE
-	    if (initial < 0 && map && inMapArea(a, px, py)) {
-		if (a->shape == SHAPE_DEFAULT) {
-		    if (initial == -n)
-			initial = -i;
-		}
-		else
-		    initial = i;
+	    l = (a->center_x - x) * (a->center_x - x)
+		+ (a->center_y - y) * (a->center_y - y);
+	    if ((min < 0 || l < min) && l < limit) {
+		n = i;
+		min = l;
 	    }
-#endif
 	}
-	else
-	    label[i] = "";
     }
-    label[n] = NULL;
-    if (initial == -n)
-	initial = 0;
-    else if (initial < 0)
-	initial *= -1;
+    return n;
+}
 
-    optionMenu(x, y, label, &selected, initial, NULL);
+static int
+searchMapArea(Buffer *buf, MapList *ml, Anchor *a_img)
+{
+    ListItem *al;
+    MapArea *a;
+    int i, n;
+    int px, py;
 
-    if (selected >= 0) {
-	for (i = 0, al = ml->area->first; al != NULL; i++, al = al->next) {
-	    if (al->ptr && i == selected)
-		return ((MapArea *) al->ptr)->url;
+    if (!(ml && ml->area && ml->area->nitem))
+	return -1;
+    if (!getMapXY(buf, a_img, &px, &py))
+	return -1;
+    n = -ml->area->nitem;
+    for (i = 0, al = ml->area->first; al != NULL; i++, al = al->next) {
+	a = (MapArea *) al->ptr;
+	if (!a)
+	    continue;
+	if (n < 0 && inMapArea(a, px, py)) {
+	    if (a->shape == SHAPE_DEFAULT) {
+		if (n == -ml->area->nitem)
+		    n = -i;
+	    }
+	    else
+		n = i;
 	}
     }
+    if (n == -ml->area->nitem)
+	return nearestMapArea(ml, px, py);
+    else if (n < 0)
+	return -n;
+    return n;
+}
+
+MapArea *
+retrieveCurrentMapArea(Buffer *buf)
+{
+    Anchor *a_img, *a_form;
+    FormItemList *fi;
+    MapList *ml;
+    ListItem *al;
+    MapArea *a;
+    int i, n;
+
+    a_img = retrieveCurrentImg(buf);
+    if (!(a_img && a_img->image && a_img->image->map))
+	return NULL;
+    a_form = retrieveCurrentForm(buf);
+    if (!(a_form && a_form->url))
+	return NULL;
+    fi = (FormItemList *)a_form->url;
+    if (!(fi && fi->parent && fi->parent->item))
+	return NULL;
+    fi = fi->parent->item;
+    ml = searchMapList(buf, fi->value ? fi->value->ptr : NULL);
+    if (!ml)
+	return NULL;
+    n = searchMapArea(buf, ml, a_img);
+    if (n < 0)
+	return NULL;
+    for (i = 0, al = ml->area->first; al != NULL; i++, al = al->next) {
+	a = (MapArea *) al->ptr;
+	if (a && i == n)
+	    return a;
+    }
     return NULL;
 }
 
-#else
-char *map1 = "<HTML><HEAD><TITLE>Image map links</TITLE></HEAD>\
-<BODY><H1>Image map links</H1>";
-
-Buffer *
-follow_map_panel(Buffer *buf, struct parsed_tagarg *arg)
-{
-    Str mappage;
-    MapList *ml;
-    ListItem *al;
-    MapArea *a;
-    char *name;
-    ParsedURL pu;
-
-    name = tag_get_value(arg, "link");
-    if (name == NULL)
-	return NULL;
-
-    for (ml = buf->maplist; ml != NULL; ml = ml->next) {
-	if (!Strcmp_charp(ml->name, name))
-	    break;
-    }
-    if (ml == NULL)
-	return NULL;
-
-    mappage = Strnew_charp(map1);
-    for (al = ml->area->first; al != NULL; al = al->next) {
-	a = (MapArea *) al->ptr;
-	if (!a)
-	    continue;
-	parseURL2(a->url, &pu, baseURL(buf));
-	Strcat_charp(mappage, "<a href=\"");
-	Strcat_charp(mappage, html_quote(parsedURL2Str(&pu)->ptr));
-	Strcat_charp(mappage, "\">");
-	Strcat_charp(mappage, html_quote(a->alt));
-	Strcat_charp(mappage, " ");
-	Strcat_charp(mappage, html_quote(a->url));
-	Strcat_charp(mappage, "</a><br>\n");
-    }
-    Strcat_charp(mappage, "</body></html>");
-
-    return loadHTMLString(mappage);
-}
-#endif
-
-#ifdef USE_IMAGE
 int
 getMapXY(Buffer *buf, Anchor *a, int *x, int *y)
 {
@@ -192,20 +180,129 @@ getMapXY(Buffer *buf, Anchor *a, int *x, int *y)
 }
 #endif
 
+Anchor *
+retrieveCurrentMap(Buffer *buf)
+{
+    Anchor *a;
+    FormItemList *fi;
+
+    a = retrieveCurrentForm(buf);
+    if (!a || !a->url)
+	return NULL;
+    fi = (FormItemList *)a->url;
+    if (fi->parent->method == FORM_METHOD_INTERNAL &&
+	!Strcmp_charp(fi->parent->action, "map"))
+	return a;
+    return NULL;
+}
+
+#if defined(USE_IMAGE) || defined(MENU_MAP)
 MapArea *
-newMapArea(char *url, char *alt, char *shape, char *coords)
+follow_map_menu(Buffer *buf, char *name, Anchor *a_img, int x, int y)
+{
+    MapList *ml;
+    ListItem *al;
+    int i, selected = -1;
+    int initial = 0;
+#ifdef MENU_MAP
+    MapArea *a;
+    char **label;
+#endif
+
+    ml = searchMapList(buf, name);
+    if (ml == NULL || ml->area == NULL || ml->area->nitem == 0)
+	return NULL;
+
+#ifdef USE_IMAGE
+    initial = searchMapArea(buf, ml, a_img);
+    if (initial < 0)
+	initial = 0;
+    else if (!image_map_list) {
+	selected = initial;
+	goto map_end;
+    }
+#endif
+
+#ifdef MENU_MAP
+    label = New_N(char *, ml->area->nitem + 1);
+    for (i = 0, al = ml->area->first; al != NULL; i++, al = al->next) {
+	a = (MapArea *) al->ptr;
+	if (a)
+	    label[i] = *a->alt ? a->alt : a->url;
+	else
+	    label[i] = "";
+    }
+    label[ml->area->nitem] = NULL;
+
+    optionMenu(x, y, label, &selected, initial, NULL);
+#endif
+
+#ifdef USE_IMAGE
+  map_end:
+#endif
+    if (selected >= 0) {
+	for (i = 0, al = ml->area->first; al != NULL; i++, al = al->next) {
+	    if (al->ptr && i == selected)
+		return (MapArea *) al->ptr;
+	}
+    }
+    return NULL;
+}
+#endif
+
+#ifndef MENU_MAP
+char *map1 = "<HTML><HEAD><TITLE>Image map links</TITLE></HEAD>\
+<BODY><H1>Image map links</H1>\
+<table>";
+
+Buffer *
+follow_map_panel(Buffer *buf, char *name)
+{
+    Str mappage;
+    MapList *ml;
+    ListItem *al;
+    MapArea *a;
+    ParsedURL pu;
+    char *p, *q;
+
+    ml = searchMapList(buf, name);
+    if (ml == NULL)
+	return NULL;
+
+    mappage = Strnew_charp(map1);
+    for (al = ml->area->first; al != NULL; al = al->next) {
+	a = (MapArea *) al->ptr;
+	if (!a)
+	    continue;
+	parseURL2(a->url, &pu, baseURL(buf));
+	p = parsedURL2Str(&pu)->ptr;
+	q = html_quote(p);
+	if (DecodeURL)
+	    p = html_quote(url_unquote_conv(p, buf->document_code));
+	else
+	    p = q;
+	Strcat_m_charp(mappage, "<tr valign=top><td><a href=\"", q, "\">",
+		       html_quote(*a->alt ? a->alt : mybasename(a->url)),
+		       "</a><td>", p, NULL);
+    }
+    Strcat_charp(mappage, "</table></body></html>");
+
+    return loadHTMLString(mappage);
+}
+#endif
+
+MapArea *
+newMapArea(char *url, char *target, char *alt, char *shape, char *coords)
 {
     MapArea *a = New(MapArea);
-#ifdef MENU_MAP
 #ifdef USE_IMAGE
     char *p;
     int i, max;
-#endif
 #endif
 
     a->url = url;
+    a->target = target;
     a->alt = alt ? alt : "";
-#ifdef MENU_MAP
 #ifdef USE_IMAGE
     a->shape = SHAPE_RECT;
     if (shape) {
@@ -222,6 +319,8 @@ newMapArea(char *url, char *alt, char *shape, char *coords)
     }
     a->coords = NULL;
     a->ncoords = 0;
+    a->center_x = 0;
+    a->center_y = 0;
     if (a->shape == SHAPE_UNKNOWN || a->shape == SHAPE_DEFAULT)
 	return a;
     if (!coords) {
@@ -240,7 +339,7 @@ newMapArea(char *url, char *alt, char *shape, char *coords)
     for (i = 0, p = coords; (a->shape == SHAPE_POLY || i < a->ncoords) && *p;) {
 	while (IS_SPACE(*p))
 	    p++;
-	if (!IS_DIGIT(*p))
+	if (!IS_DIGIT(*p) && *p != '-' && *p != '+')
 	    break;
 	if (a->shape == SHAPE_POLY) {
 	    if (max <= i) {
@@ -251,6 +350,8 @@ newMapArea(char *url, char *alt, char *shape, char *coords)
 	}
 	a->coords[i] = (short)atoi(p);
 	i++;
+	if (*p == '-' || *p == '+')
+	    p++;
 	while (IS_DIGIT(*p))
 	    p++;
 	if (*p != ',' && !IS_SPACE(*p))
@@ -264,17 +365,104 @@ newMapArea(char *url, char *alt, char *shape, char *coords)
 	a->shape = SHAPE_UNKNOWN;
 	a->coords = NULL;
 	a->ncoords = 0;
+	return a;
     }
     if (a->shape == SHAPE_POLY) {
 	a->ncoords = a->ncoords / 2 * 2;
 	a->coords[a->ncoords] = a->coords[0];
 	a->coords[a->ncoords + 1] = a->coords[1];
     }
-#endif
+    if (a->shape == SHAPE_CIRCLE) {
+	a->center_x = a->coords[0];
+	a->center_y = a->coords[1];
+    }
+    else {
+	for (i = 0; i < a->ncoords / 2; i++) {
+	    a->center_x += a->coords[2 * i];
+	    a->center_y += a->coords[2 * i + 1];
+	}
+	a->center_x /= a->ncoords / 2;
+	a->center_y /= a->ncoords / 2;
+    }
 #endif
     return a;
 }
 
+/* append image map links */
+static void
+append_map_info(Buffer *buf, Str tmp, FormItemList *fi)
+{
+    MapList *ml;
+    ListItem *al;
+    MapArea *a;
+    ParsedURL pu;
+    char *p, *q;
+
+    ml = searchMapList(buf, fi->value ? fi->value->ptr : NULL);
+    if (ml == NULL)
+	return;
+
+    Strcat_m_charp(tmp,
+		   "<tr valign=top><td colspan=2>Links of current image map",
+		   "<tr valign=top><td colspan=2><table>", NULL);
+    for (al = ml->area->first; al != NULL; al = al->next) {
+	a = (MapArea *) al->ptr;
+	if (!a)
+	    continue;
+	parseURL2(a->url, &pu, baseURL(buf));
+	q = html_quote(parsedURL2Str(&pu)->ptr);
+	if (DecodeURL)
+	    p = html_quote(url_unquote_conv(a->url, buf->document_code));
+	else
+	    p = html_quote(a->url);
+	Strcat_m_charp(tmp, "<tr valign=top><td>&nbsp;&nbsp;<td><a href=\"",
+		       q, "\">",
+		       html_quote(*a->alt ? a->alt : mybasename(a->url)),
+		       "</a><td>", p, "\n", NULL);
+    }
+    Strcat_charp(tmp, "</table>");
+}
+
+/* append links */
+static void
+append_link_info(Buffer *buf, Str html, LinkList * link)
+{
+    LinkList *l;
+    ParsedURL pu;
+    char *url;
+
+    if (!link)
+	return;
+
+    Strcat_charp(html, "<hr width=50%><h1>Link information</h1><table>\n");
+    for (l = link; l; l = l->next) {
+	if (l->url) {
+	    parseURL2(l->url, &pu, baseURL(buf));
+	    url = html_quote(parsedURL2Str(&pu)->ptr);
+	}
+	else
+	    url = "(empty)";
+	Strcat_m_charp(html, "<tr valign=top><td><a href=\"", url, "\">",
+		       l->title ? html_quote(l->title) : "(empty)", "</a><td>",
+		       NULL);
+	if (l->type == LINK_TYPE_REL)
+	    Strcat_charp(html, "[Rel]");
+	else if (l->type == LINK_TYPE_REV)
+	    Strcat_charp(html, "[Rev]");
+	if (!l->url)
+	    url = "(empty)";
+	else if (DecodeURL)
+	    url = html_quote(url_unquote_conv(l->url, buf->document_code));
+	else
+	    url = html_quote(l->url);
+	Strcat_m_charp(html, "<td>", url, NULL);
+	if (l->ctype)
+	    Strcat_m_charp(html, " (", html_quote(l->ctype), ")", NULL);
+	Strcat_charp(html, "\n");
+    }
+    Strcat_charp(html, "</table>\n");
+}
+
 /* append frame URL */
 static void
 append_frame_info(Buffer *buf, Str html, struct frameset *set, int level)
@@ -293,27 +481,27 @@ append_frame_info(Buffer *buf, Str html, struct frameset *set, int level)
 	    case F_BODY:
 		if (frame.body->url == NULL)
 		    break;
+		Strcat_charp(html, "<pre_int>");
 		for (j = 0; j < level; j++)
 		    Strcat_charp(html, "   ");
 		q = html_quote(frame.body->url);
-		Strcat_charp(html, "<a href=\"");
-		Strcat_charp(html, q);
-		Strcat_charp(html, "\">");
+		Strcat_m_charp(html, "<a href=\"", q, "\">", NULL);
 		if (frame.body->name) {
-		    p = file_unquote(frame.body->name);
-#ifdef JP_CHARSET
-		    p = conv(p, buf->document_code, InnerCode)->ptr;
-#endif
-		    p = html_quote(p);
+		    p = html_quote(url_unquote_conv(frame.body->name,
+						    buf->document_code));
 		    Strcat_charp(html, p);
 		}
-		Strcat_charp(html, " ");
-		Strcat_charp(html, q);
-		Strcat_charp(html, "</a>\n");
+		if (DecodeURL)
+		    p = html_quote(url_unquote_conv(frame.body->url,
+						    buf->document_code));
+		else
+		    p = q;
+		Strcat_m_charp(html, " ", p, "</a></pre_int><br>\n", NULL);
 #ifdef USE_SSL
 		if (frame.body->ssl_certificate)
-		    Strcat_m_charp(html, "<blockquote><pre>\n",
-				   frame.body->ssl_certificate,
+		    Strcat_m_charp(html,
+				   "<blockquote><h2>SSL certificate</h2><pre>\n",
+				   html_quote(frame.body->ssl_certificate),
 				   "</pre></blockquote>\n", NULL);
 #endif
 		break;
@@ -333,76 +521,95 @@ page_info_panel(Buffer *buf)
 {
     Str tmp = Strnew_size(1024);
     Anchor *a;
-    Str s;
     ParsedURL pu;
     TextListItem *ti;
     struct frameset *f_set = NULL;
     int all;
+    char *p, *q;
 
-    Strcat_charp(tmp,
-		 "<html><head><title>Information about current page</title></head><body>");
-    Strcat_charp(tmp,
-		 "<h1>Information about current page</h1><table cellpadding=0>");
+    Strcat_charp(tmp, "<html><head>\
+<title>Information about current page</title>\
+</head><body>\
+<h1>Information about current page</h1>\n");
     if (buf == NULL)
 	goto end;
-    Strcat_charp(tmp, "<tr><td nowrap>Title<td>");
-    Strcat_charp(tmp, html_quote(buf->buffername));
-    Strcat_charp(tmp, "<tr><td nowrap>Current URL<td>");
-    Strcat_charp(tmp, html_quote(parsedURL2Str(&buf->currentURL)->ptr));
-    Strcat_charp(tmp, "<tr><td nowrap>Document Type<td>");
-    if (buf->real_type)
-	Strcat_charp(tmp, buf->real_type);
-    else
-	Strcat_charp(tmp, "unknown");
-    Strcat_charp(tmp, "<tr><td nowrap>Last Modified<td>");
-    Strcat_charp(tmp, html_quote(last_modified(buf)));
-#ifdef JP_CHARSET
-    Strcat_charp(tmp, "<tr><td nowrap>Document Code<td>");
-    Strcat_charp(tmp, code_to_str(buf->document_code));
-#endif				/* JP_CHARSET */
-    Strcat_charp(tmp, "<tr><td nowrap>Number of lines<td>");
     all = buf->allLine;
     if (all == 0 && buf->lastLine)
 	all = buf->lastLine->linenumber;
-    Strcat(tmp, Sprintf("%d", all));
-    Strcat_charp(tmp, "<tr><td nowrap>Transferred bytes<td>");
-    Strcat(tmp, Sprintf("%d", buf->trbyte));
+    p = parsedURL2Str(&buf->currentURL)->ptr;
+    if (DecodeURL)
+	p = url_unquote_conv(p, 0);
+    Strcat_m_charp(tmp, "<table cellpadding=0>",
+		   "<tr valign=top><td nowrap>Title<td>",
+		   html_quote(buf->buffername),
+		   "<tr valign=top><td nowrap>Current URL<td>",
+		   html_quote(p),
+		   "<tr valign=top><td nowrap>Document Type<td>",
+		   buf->real_type ? html_quote(buf->real_type) : "unknown",
+		   "<tr valign=top><td nowrap>Last Modified<td>",
+		   html_quote(last_modified(buf)),
+#ifdef JP_CHARSET
+		   "<tr valign=top><td nowrap>Document Code<td>",
+		   code_to_str(buf->document_code),
+#endif				/* JP_CHARSET */
+		   "<tr valign=top><td nowrap>Number of lines<td>",
+		   Sprintf("%d", all)->ptr,
+		   "<tr valign=top><td nowrap>Transferred bytes<td>",
+		   Sprintf("%d", buf->trbyte)->ptr, NULL);
 
     a = retrieveCurrentAnchor(buf);
     if (a != NULL) {
-	char *aurl;
 	parseURL2(a->url, &pu, baseURL(buf));
-	s = parsedURL2Str(&pu);
-	aurl = html_quote(s->ptr);
-	Strcat_charp(tmp, "<tr><td nowrap>URL of current anchor<td>");
-	Strcat_m_charp(tmp, "<a href=\"", aurl, "\">", aurl, "</a>", NULL);
+	p = parsedURL2Str(&pu)->ptr;
+	q = html_quote(p);
+	if (DecodeURL)
+	    p = html_quote(url_unquote_conv(p, buf->document_code));
+	else
+	    p = q;
+	Strcat_m_charp(tmp,
+		       "<tr valign=top><td nowrap>URL of current anchor<td><a href=\"",
+		       q, "\">", p, "</a>", NULL);
     }
     a = retrieveCurrentImg(buf);
     if (a != NULL) {
 	parseURL2(a->url, &pu, baseURL(buf));
-	s = parsedURL2Str(&pu);
-	Strcat_charp(tmp, "<tr><td nowrap>URL of current image<td>");
-	Strcat_charp(tmp, "<a href=\"");
-	Strcat_charp(tmp, html_quote(s->ptr));
-	Strcat_charp(tmp, "\">");
-	Strcat_charp(tmp, html_quote(s->ptr));
-	Strcat_charp(tmp, "</a>");
+	p = parsedURL2Str(&pu)->ptr;
+	q = html_quote(p);
+	if (DecodeURL)
+	    p = html_quote(url_unquote_conv(p, buf->document_code));
+	else
+	    p = q;
+	Strcat_m_charp(tmp,
+		       "<tr valign=top><td nowrap>URL of current image<td><a href=\"",
+		       q, "\">", p, "</a>", NULL);
     }
     a = retrieveCurrentForm(buf);
     if (a != NULL) {
-	s = Strnew_charp(form2str((FormItemList *)a->url));
-	Strcat_charp(tmp, "<tr><td nowrap>Method/type of current form<td>");
-	Strcat_charp(tmp, html_quote(s->ptr));
+	FormItemList *fi = (FormItemList *)a->url;
+	p = form2str(fi);
+	if (DecodeURL)
+	    p = html_quote(url_unquote_conv(p, buf->document_code));
+	else
+	    p = html_quote(p);
+	Strcat_m_charp(tmp,
+		       "<tr valign=top><td nowrap>Method/type of current form&nbsp;<td>",
+		       p, NULL);
+	if (fi->parent->method == FORM_METHOD_INTERNAL
+	    && !Strcmp_charp(fi->parent->action, "map"))
+	    append_map_info(buf, tmp, fi->parent->item);
     }
     Strcat_charp(tmp, "</table>\n");
+
+    append_link_info(buf, tmp, buf->linklist);
+
     if (buf->document_header != NULL) {
-	Strcat_charp(tmp, "<hr width=50%>\n");
-	Strcat_charp(tmp, "<h1>Header information</h1>\n");
-	for (ti = buf->document_header->first; ti != NULL; ti = ti->next) {
-	    Strcat_charp(tmp, html_quote(ti->ptr));
-	    Strcat_charp(tmp, "<br>");
-	}
+	Strcat_charp(tmp, "<hr width=50%><h1>Header information</h1><pre>\n");
+	for (ti = buf->document_header->first; ti != NULL; ti = ti->next)
+	    Strcat_m_charp(tmp, "<pre_int>", html_quote(ti->ptr),
+			   "</pre_int>\n", NULL);
+	Strcat_charp(tmp, "</pre>\n");
     }
+
     if (buf->frameset != NULL)
 	f_set = buf->frameset;
     else if (buf->bufferprop & BP_FRAME &&
@@ -410,17 +617,13 @@ page_info_panel(Buffer *buf)
 	f_set = buf->nextBuffer->frameset;
 
     if (f_set) {
-	Strcat_charp(tmp, "<hr width=50%><h1>Frame information</h1><pre>");
+	Strcat_charp(tmp, "<hr width=50%><h1>Frame information</h1>\n");
 	append_frame_info(buf, tmp, f_set, 0);
-	Strcat_charp(tmp, "</pre>");
     }
 #ifdef USE_SSL
-    if (buf->ssl_certificate == NULL)
-	goto end;
-    Strcat_charp(tmp, "<h1>SSL certificate</h1>\n");
-    Strcat_charp(tmp, "<pre>\n");
-    Strcat_charp(tmp, buf->ssl_certificate);
-    Strcat_charp(tmp, "</pre>\n");
+    if (buf->ssl_certificate)
+	Strcat_m_charp(tmp, "<h1>SSL certificate</h1><pre>\n",
+		       html_quote(buf->ssl_certificate), "</pre>\n", NULL);
 #endif
   end:
     Strcat_charp(tmp, "</body></html>");
