@@ -1,4 +1,4 @@
-/* $Id: menu.c,v 1.15 2002/03/19 16:06:52 ukai Exp $ */
+/* $Id: menu.c,v 1.31 2003/02/05 16:44:00 ukai Exp $ */
 /* 
  * w3m menu.c
  */
@@ -85,6 +85,8 @@ static int mNext(char c);
 static int mPrev(char c);
 static int mFore(char c);
 static int mBack(char c);
+static int mLineU(char c);
+static int mLineD(char c);
 static int mOk(char c);
 static int mCancel(char c);
 static int mClose(char c);
@@ -109,7 +111,7 @@ static int (*MenuKeymap[128]) (char c) = {
 /*  C-h     C-i     C-j     C-k     C-l     C-m     C-n     C-o      */
     mCancel,mNull,  mOk,    mNull,  mNull,  mOk,    mDown,  mNull,
 /*  C-p     C-q     C-r     C-s     C-t     C-u     C-v     C-w      */
-    mUp,    mNull,  mNull,  mNull,  mNull,  mNull,  mNext,  mNull,
+    mUp,    mNull,  mSrchB, mSrchF, mNull,  mNull,  mNext,  mNull,
 /*  C-x     C-y     C-z     C-[     C-\     C-]     C-^     C-_      */
     mNull,  mNull,  mSusp,  mEsc,   mNull,  mNull,  mNull,  mNull,
 /*  SPC     !       "       #       $       %       &       '        */
@@ -123,7 +125,7 @@ static int (*MenuKeymap[128]) (char c) = {
 /*  @       A       B       C       D       E       F       G        */
     mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
 /*  H       I       J       K       L       M       N       O        */
-    mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mSrchP, mNull,
+    mNull,  mNull,  mLineU, mLineD, mNull,  mNull,  mSrchP, mNull,
 /*  P       Q       R       S       T       U       V       W        */
     mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,  mNull,
 /*  X       Y       Z       [       \       ]       ^       _        */
@@ -276,6 +278,16 @@ static int smDelBuf(char c);
 
 /* --- SelectMenu (END) --- */
 
+/* --- SelTabMenu --- */
+
+static Menu SelTabMenu;
+static int SelTabV = 0;
+static void initSelTabMenu(void);
+static void smChTab(void);
+static int smDelTab(char c);
+
+/* --- SelTabMenu (END) --- */
+
 /* --- MainMenu --- */
 
 static Menu MainMenu;
@@ -284,12 +296,14 @@ static MenuItem MainMenuItem[] = {
     /* type        label         variabel value func     popup keys data  */
     {MENU_FUNC, "         (b)", NULL, 0, backBf, NULL, "b", NULL},
     {MENU_POPUP, "Хåե (s)", NULL, 0, NULL, &SelectMenu, "s", NULL},
+    {MENU_POPUP, "     (t)", NULL, 0, NULL, &SelTabMenu, "tT", NULL},
     {MENU_FUNC, "ɽ (v)", NULL, 0, vwSrc, NULL, "vV", NULL},
     {MENU_FUNC, "Խ (e)", NULL, 0, editBf, NULL, "eE", NULL},
     {MENU_FUNC, "¸ (S)", NULL, 0, svSrc, NULL, "S", NULL},
     {MENU_FUNC, "ɤ߹   (r)", NULL, 0, reload, NULL, "rR", NULL},
     {MENU_NOP, "", NULL, 0, nulcmd, NULL, "", NULL},
     {MENU_FUNC, "󥯤ɽ (a)", NULL, 0, followA, NULL, "a", NULL},
+    {MENU_FUNC, "֤ɽ (n)", NULL, 0, tabA, NULL, "nN", NULL},
     {MENU_FUNC, "󥯤¸ (A)", NULL, 0, svA, NULL, "A", NULL},
     {MENU_FUNC, "ɽ   (i)", NULL, 0, followI, NULL, "i", NULL},
     {MENU_FUNC, "¸   (I)", NULL, 0, svI, NULL, "I", NULL},
@@ -307,12 +321,14 @@ static MenuItem MainMenuItem[] = {
     /* type        label           variable value func     popup keys data  */
     {MENU_FUNC, " Back         (b) ", NULL, 0, backBf, NULL, "b", NULL},
     {MENU_POPUP, " Select Buffer(s) ", NULL, 0, NULL, &SelectMenu, "s", NULL},
+    {MENU_POPUP, " Select Tab   (t) ", NULL, 0, NULL, &SelTabMenu, "tT", NULL},
     {MENU_FUNC, " View Source  (v) ", NULL, 0, vwSrc, NULL, "vV", NULL},
     {MENU_FUNC, " Edit Source  (e) ", NULL, 0, editBf, NULL, "eE", NULL},
     {MENU_FUNC, " Save Source  (S) ", NULL, 0, svSrc, NULL, "S", NULL},
     {MENU_FUNC, " Reload       (r) ", NULL, 0, reload, NULL, "rR", NULL},
     {MENU_NOP, " ---------------- ", NULL, 0, nulcmd, NULL, "", NULL},
     {MENU_FUNC, " Go Link      (a) ", NULL, 0, followA, NULL, "a", NULL},
+    {MENU_FUNC, "   on New Tab (n) ", NULL, 0, tabA, NULL, "nN", NULL},
     {MENU_FUNC, " Save Link    (A) ", NULL, 0, svA, NULL, "A", NULL},
     {MENU_FUNC, " View Image   (i) ", NULL, 0, followI, NULL, "i", NULL},
     {MENU_FUNC, " Save Image   (I) ", NULL, 0, svI, NULL, "I", NULL},
@@ -882,6 +898,42 @@ mBack(char c)
     return (MENU_NOTHING);
 }
 
+static int
+mLineU(char c)
+{
+    int mselect = CurrentMenu->select;
+
+    if (mselect >= CurrentMenu->nitem)
+	return mLast(c);
+    if (CurrentMenu->offset + CurrentMenu->height >= CurrentMenu->nitem)
+	mselect++;
+    else {
+	down_menu(CurrentMenu, 1);
+	if (mselect < CurrentMenu->offset)
+	    mselect++;
+    }
+    goto_menu(CurrentMenu, mselect, 1);
+    return (MENU_NOTHING);
+}
+
+static int
+mLineD(char c)
+{
+    int mselect = CurrentMenu->select;
+
+    if (mselect <= 0)
+	return mTop(c);
+    if (CurrentMenu->offset <= 0)
+	mselect--;
+    else {
+	up_menu(CurrentMenu, 1);
+	if (mselect >= CurrentMenu->offset + CurrentMenu->height)
+	    mselect--;
+    }
+    goto_menu(CurrentMenu, mselect, -1);
+    return (MENU_NOTHING);
+}
+
 static int
 mOk(char c)
 {
@@ -1071,13 +1123,23 @@ mSrchP(char c)
 #define MOUSE_BTN5_DOWN_XTERM 65
 #define MOUSE_BTN_UP 3
 #define MOUSE_BTN_RESET -1
-#define MOUSE_SCROLL_LINE 5
+
+static int
+mMouse_scroll_line(void)
+{
+    int i = 0;
+    if (relative_wheel_scroll)
+	i = (relative_wheel_scroll_ratio * CurrentMenu->height + 99) / 100;
+    else
+	i = fixed_wheel_scroll_count;
+    return i ? i : 1;
+}
 
 static int
 process_mMouse(int btn, int x, int y)
 {
     Menu *menu;
-    int mselect;
+    int mselect, i;
     static int press_btn = MOUSE_BTN_RESET, press_x, press_y;
     char c = ' ';
 
@@ -1087,7 +1149,9 @@ process_mMouse(int btn, int x, int y)
 	return (MENU_NOTHING);
 
     if (btn == MOUSE_BTN_UP) {
-	if (press_btn == MOUSE_BTN1_DOWN || press_btn == MOUSE_BTN3_DOWN) {
+	switch (press_btn) {
+	case MOUSE_BTN1_DOWN:
+	case MOUSE_BTN3_DOWN:
 	    if (x < menu->x - FRAME_WIDTH ||
 		x >= menu->x + menu->width + FRAME_WIDTH ||
 		y < menu->y - 1 || y >= menu->y + menu->height + 1) {
@@ -1099,6 +1163,16 @@ process_mMouse(int btn, int x, int y)
 		      x < menu->x + menu->width + FRAME_WIDTH)) {
 		return (MENU_NOTHING);
 	    }
+	    else if (press_y > y) {
+		for (i = 0; i < press_y - y; i++)
+		    mLineU(c);
+		return (MENU_NOTHING);
+	    }
+	    else if (press_y < y) {
+		for (i = 0; i < y - press_y; i++)
+		    mLineD(c);
+		return (MENU_NOTHING);
+	    }
 	    else if (y == menu->y - 1) {
 		mPrev(c);
 		return (MENU_NOTHING);
@@ -1113,13 +1187,34 @@ process_mMouse(int btn, int x, int y)
 		    return (MENU_NOTHING);
 		return (select_menu(menu, mselect));
 	    }
+	    break;
+	case MOUSE_BTN4_DOWN_RXVT:
+	    for (i = 0; i < mMouse_scroll_line(); i++)
+		mLineD(c);
+	    break;
+	case MOUSE_BTN5_DOWN_RXVT:
+	    for (i = 0; i < mMouse_scroll_line(); i++)
+		mLineU(c);
+	    break;
 	}
     }
-    else {
+    else if (btn == MOUSE_BTN4_DOWN_XTERM) {
+	for (i = 0; i < mMouse_scroll_line(); i++)
+	    mLineD(c);
+    }
+    else if (btn == MOUSE_BTN5_DOWN_XTERM) {
+	for (i = 0; i < mMouse_scroll_line(); i++)
+	    mLineU(c);
+    }
+
+    if (btn != MOUSE_BTN4_DOWN_RXVT || press_btn == MOUSE_BTN_RESET) {
 	press_btn = btn;
 	press_x = x;
 	press_y = y;
     }
+    else {
+	press_btn = MOUSE_BTN_RESET;
+    }
     return (MENU_NOTHING);
 }
 
@@ -1129,6 +1224,14 @@ mMouse(char c)
     int btn, x, y;
 
     btn = (unsigned char)getch() - 32;
+#if defined(__CYGWIN__)
+    if (cygwin_mouse_btn_swapped) {
+	if (btn == MOUSE_BTN2_DOWN)
+	    btn = MOUSE_BTN3_DOWN;
+	else if (btn == MOUSE_BTN3_DOWN)
+	    btn = MOUSE_BTN2_DOWN;
+    }
+#endif
     x = (unsigned char)getch() - 33;
     if (x < 0)
 	x += 0x100;
@@ -1209,9 +1312,10 @@ void
 popupMenu(int x, int y, Menu *menu)
 {
     initSelectMenu();
+    initSelTabMenu();
 
     menu->cursorX = Currentbuf->cursorX + Currentbuf->rootX;
-    menu->cursorY = Currentbuf->cursorY;
+    menu->cursorY = Currentbuf->cursorY + Currentbuf->rootY;
     menu->x = x + FRAME_WIDTH + 1;
     menu->y = y + 2;
 
@@ -1230,6 +1334,8 @@ mainMn(void)
     Menu *menu = &MainMenu;
     char *data;
     int n;
+    int x = Currentbuf->cursorX + Currentbuf->rootX,
+	y = Currentbuf->cursorY + Currentbuf->rootY;
 
     data = searchKeyData();
     if (data != NULL) {
@@ -1238,8 +1344,13 @@ mainMn(void)
 	    return;
 	menu = w3mMenuList[n].menu;
     }
-    popupMenu(Currentbuf->cursorX + Currentbuf->rootX, Currentbuf->cursorY,
-	      menu);
+#ifdef USE_MOUSE
+    if (mouse_action.in_action) {
+	x = mouse_action.cursorX;
+	y = mouse_action.cursorY;
+    }
+#endif
+    popupMenu(x, y, menu);
 }
 
 /* --- MainMenu (END) --- */
@@ -1249,7 +1360,16 @@ mainMn(void)
 void
 selMn(void)
 {
-    popupMenu(Currentbuf->cursorX, Currentbuf->cursorY, &SelectMenu);
+    int x = Currentbuf->cursorX + Currentbuf->rootX,
+	y = Currentbuf->cursorY + Currentbuf->rootY;
+
+#ifdef USE_MOUSE
+    if (mouse_action.in_action) {
+	x = mouse_action.cursorX;
+	y = mouse_action.cursorY;
+    }
+#endif
+    popupMenu(x, y, &SelectMenu);
 }
 
 static void
@@ -1259,6 +1379,7 @@ initSelectMenu(void)
     Buffer *buf;
     Str str;
     char **label;
+    char *p;
     static char *comment = " SPC for select / D for delete buffer ";
 
     SelectV = -1;
@@ -1274,7 +1395,6 @@ initSelectMenu(void)
 	if (buf->filename != NULL) {
 	    switch (buf->currentURL.scheme) {
 	    case SCM_LOCAL:
-	    case SCM_LOCAL_CGI:
 		if (strcmp(buf->currentURL.file, "-")) {
 		    Strcat_char(str, ' ');
 		    Strcat_charp(str,
@@ -1286,7 +1406,10 @@ initSelectMenu(void)
 		break;
 	    default:
 		Strcat_char(str, ' ');
-		Strcat(str, parsedURL2Str(&buf->currentURL));
+		p = parsedURL2Str(&buf->currentURL)->ptr;
+		if (DecodeURL)
+		    p = url_unquote_conv(p, 0);
+		Strcat_charp(str, p);
 		break;
 	    }
 	}
@@ -1312,7 +1435,7 @@ initSelectMenu(void)
     new_option_menu(&SelectMenu, label, &SelectV, smChBuf);
     SelectMenu.initial = SelectV;
     SelectMenu.cursorX = Currentbuf->cursorX + Currentbuf->rootX;
-    SelectMenu.cursorY = Currentbuf->cursorY;
+    SelectMenu.cursorY = Currentbuf->cursorY + Currentbuf->rootY;
     SelectMenu.keymap['D'] = smDelBuf;
     SelectMenu.item[nitem].type = MENU_NOP;
 }
@@ -1380,6 +1503,150 @@ smDelBuf(char c)
 
 /* --- SelectMenu (END) --- */
 
+/* --- SelTabMenu --- */
+
+void
+tabMn(void)
+{
+    int x = Currentbuf->cursorX + Currentbuf->rootX,
+	y = Currentbuf->cursorY + Currentbuf->rootY;
+
+#ifdef USE_MOUSE
+    if (mouse_action.in_action) {
+	x = mouse_action.cursorX;
+	y = mouse_action.cursorY;
+    }
+#endif
+    popupMenu(x, y, &SelTabMenu);
+}
+
+static void
+initSelTabMenu(void)
+{
+    int i, nitem, len = 0, l;
+    TabBuffer *tab;
+    Buffer *buf;
+    Str str;
+    char **label;
+    char *p;
+    static char *comment = " SPC for select / D for delete tab ";
+
+    SelTabV = -1;
+    for (i = 0, tab = LastTab; tab != NULL; i++, tab = tab->prevTab) {
+	if (tab == CurrentTab)
+	    SelTabV = i;
+    }
+    nitem = i;
+
+    label = New_N(char *, nitem + 2);
+    for (i = 0, tab = LastTab; i < nitem; i++, tab = tab->prevTab) {
+	buf = tab->currentBuffer;
+	str = Sprintf("<%s>", buf->buffername);
+	if (buf->filename != NULL) {
+	    switch (buf->currentURL.scheme) {
+	    case SCM_LOCAL:
+		if (strcmp(buf->currentURL.file, "-")) {
+		    Strcat_char(str, ' ');
+		    Strcat_charp(str,
+				 conv_from_system(buf->currentURL.real_file));
+		}
+		break;
+		/* case SCM_UNKNOWN: */
+	    case SCM_MISSING:
+		break;
+	    default:
+		p = parsedURL2Str(&buf->currentURL)->ptr;
+		if (DecodeURL)
+		    p = url_unquote_conv(p, 0);
+		Strcat_charp(str, p);
+		break;
+	    }
+	}
+	label[i] = str->ptr;
+	if (len < str->length)
+	    len = str->length;
+    }
+    l = strlen(comment);
+    if (len < l + 4)
+	len = l + 4;
+    if (len > COLS - 2 * FRAME_WIDTH)
+	len = COLS - 2 * FRAME_WIDTH;
+    len = (len > 1) ? ((len - l + 1) / 2) : 0;
+    str = Strnew();
+    for (i = 0; i < len; i++)
+	Strcat_char(str, '-');
+    Strcat_charp(str, comment);
+    for (i = 0; i < len; i++)
+	Strcat_char(str, '-');
+    label[nitem] = str->ptr;
+    label[nitem + 1] = NULL;
+
+    new_option_menu(&SelTabMenu, label, &SelTabV, smChTab);
+    SelTabMenu.initial = SelTabV;
+    SelTabMenu.cursorX = Currentbuf->cursorX + Currentbuf->rootX;
+    SelTabMenu.cursorY = Currentbuf->cursorY + Currentbuf->rootY;
+    SelTabMenu.keymap['D'] = smDelTab;
+    SelTabMenu.item[nitem].type = MENU_NOP;
+}
+
+static void
+smChTab(void)
+{
+    int i;
+    TabBuffer *tab;
+    Buffer *buf;
+
+    if (SelTabV < 0 || SelTabV >= SelTabMenu.nitem)
+	return;
+    for (i = 0, tab = LastTab; i < SelTabV && tab != NULL;
+	 i++, tab = tab->prevTab) ;
+    CurrentTab = tab;
+    for (tab = LastTab; tab != NULL; tab = tab->prevTab) {
+	if (tab == CurrentTab)
+	    continue;
+	buf = tab->currentBuffer;
+#ifdef USE_IMAGE
+	deleteImage(buf);
+#endif
+	if (clear_buffer)
+	    tmpClearBuffer(buf);
+    }
+}
+
+static int
+smDelTab(char c)
+{
+    int i, x, y, mselect;
+    TabBuffer *tab;
+
+    if (CurrentMenu->select < 0 || CurrentMenu->select >= SelTabMenu.nitem)
+	return (MENU_NOTHING);
+    for (i = 0, tab = LastTab; i < CurrentMenu->select && tab != NULL;
+	 i++, tab = tab->prevTab) ;
+    deleteTab(tab);
+
+    x = CurrentMenu->x;
+    y = CurrentMenu->y;
+    mselect = CurrentMenu->select;
+
+    initSelTabMenu();
+
+    CurrentMenu->x = x;
+    CurrentMenu->y = y;
+
+    geom_menu(CurrentMenu, x, y, 0);
+
+    CurrentMenu->select = (mselect <= CurrentMenu->nitem - 2) ? mselect
+	: (CurrentMenu->nitem - 2);
+
+    displayBuffer(Currentbuf, B_FORCE_REDRAW);
+    draw_all_menu(CurrentMenu);
+    select_menu(CurrentMenu, CurrentMenu->select);
+    return (MENU_NOTHING);
+}
+
+/* --- SelectMenu (END) --- */
+
 /* --- OptionMenu --- */
 
 void
@@ -1419,7 +1686,10 @@ initMenu(void)
     w3mMenuList[1].id = "Select";
     w3mMenuList[1].menu = &SelectMenu;
     w3mMenuList[1].item = NULL;
-    w3mMenuList[2].id = NULL;
+    w3mMenuList[2].id = "SelectTab";
+    w3mMenuList[2].menu = &SelTabMenu;
+    w3mMenuList[2].item = NULL;
+    w3mMenuList[3].id = NULL;
 
     if ((mf = fopen(rcFile(MENU_FILE), "rt")) == NULL)
 	goto create_menu;
@@ -1554,4 +1824,253 @@ getMenuN(MenuList *list, char *id)
 
 /* --- InitMenu (END) --- */
 
+LinkList *
+link_menu(Buffer *buf)
+{
+    Menu menu;
+    LinkList *l;
+    int i, nitem, len = 0, linkV = -1;
+    char **label;
+    Str str;
+    char *p;
+
+    if (!buf->linklist)
+	return NULL;
+
+    for (i = 0, l = buf->linklist; l; i++, l = l->next) ;
+    nitem = i;
+
+    label = New_N(char *, nitem + 1);
+    for (i = 0, l = buf->linklist; l; i++, l = l->next) {
+	str = Strnew_charp(l->title ? l->title : "(empty)");
+	if (l->type == LINK_TYPE_REL)
+	    Strcat_charp(str, " [Rel] ");
+	else if (l->type == LINK_TYPE_REV)
+	    Strcat_charp(str, " [Rev] ");
+	else
+	    Strcat_charp(str, " ");
+	if (!l->url)
+	    p = "";
+	else if (DecodeURL)
+	    p = url_unquote_conv(l->url, buf->document_code);
+	else
+	    p = l->url;
+	Strcat_charp(str, p);
+	label[i] = str->ptr;
+	if (len < str->length)
+	    len = str->length;
+    }
+    label[nitem] = NULL;
+
+    new_option_menu(&menu, label, &linkV, NULL);
+
+    menu.initial = 0;
+    menu.cursorX = buf->cursorX + buf->rootX;
+    menu.cursorY = buf->cursorY + buf->rootY;
+    menu.x = menu.cursorX + FRAME_WIDTH + 1;
+    menu.y = menu.cursorY + 2;
+
+    popup_menu(NULL, &menu);
+
+    if (linkV < 0)
+	return NULL;
+    for (i = 0, l = buf->linklist; l; i++, l = l->next) {
+	if (i == linkV)
+	    return l;
+    }
+    return NULL;
+}
+
+/* --- LinkMenu (END) --- */
+
+Anchor *
+accesskey_menu(Buffer *buf)
+{
+    Menu menu;
+    AnchorList *al = buf->href;
+    Anchor *a;
+    Anchor **ap;
+    int i, n, nitem = 0, key = -1;
+    char **label;
+    char *t;
+    unsigned char c;
+
+    if (!al)
+	return NULL;
+    for (i = 0; i < al->nanchor; i++) {
+	a = &al->anchors[i];
+	if (!a->slave && a->accesskey && IS_ASCII(a->accesskey))
+	    nitem++;
+    }
+    if (!nitem)
+	return NULL;
+
+    label = New_N(char *, nitem + 1);
+    ap = New_N(Anchor *, nitem);
+    for (i = 0, n = 0; i < al->nanchor; i++) {
+	a = &al->anchors[i];
+	if (!a->slave && a->accesskey && IS_ASCII(a->accesskey)) {
+	    t = getAnchorText(buf, al, a);
+	    label[n] = Sprintf("%c: %s", a->accesskey, t ? t : "")->ptr;
+	    ap[n] = a;
+	    n++;
+	}
+    }
+    label[nitem] = NULL;
+
+    new_option_menu(&menu, label, &key, NULL);
+
+    menu.initial = 0;
+    menu.cursorX = buf->cursorX + buf->rootX;
+    menu.cursorY = buf->cursorY + buf->rootY;
+    menu.x = menu.cursorX + FRAME_WIDTH + 1;
+    menu.y = menu.cursorY + 2;
+    for (i = 0; i < 128; i++)
+	menu.keyselect[i] = -1;
+    for (i = 0; i < nitem; i++) {
+	c = ap[i]->accesskey;
+	menu.keymap[(int)c] = mSelect;
+	menu.keyselect[(int)c] = i;
+    }
+    for (i = 0; i < nitem; i++) {
+	c = ap[i]->accesskey;
+	if (!IS_ALPHA(c) || menu.keyselect[n] >= 0)
+	    continue;
+	c = TOLOWER(c);
+	menu.keymap[(int)c] = mSelect;
+	menu.keyselect[(int)c] = i;
+	c = TOUPPER(c);
+	menu.keymap[(int)c] = mSelect;
+	menu.keyselect[(int)c] = i;
+    }
+
+    a = retrieveCurrentAnchor(buf);
+    if (a && a->accesskey && IS_ASCII(a->accesskey)) {
+	for (i = 0; i < nitem; i++) {
+	    if (a->hseq == ap[i]->hseq) {
+		menu.initial = i;
+		break;
+	    }
+	}
+    }
+
+    popup_menu(NULL, &menu);
+
+    return (key >= 0) ? ap[key] : NULL;
+}
+
+static char lmKeys[] = "abcdefgimopqrstuvwxyz";
+static char lmKeys2[] = "1234567890ABCDEFGHILMOPQRSTUVWXYZ";
+#define nlmKeys (sizeof(lmKeys) - 1)
+#define nlmKeys2 (sizeof(lmKeys2) - 1)
+
+static int
+lmGoto(char c)
+{
+    if (IS_ASCII(c) && CurrentMenu->keyselect[(int)c] >= 0) {
+	goto_menu(CurrentMenu, CurrentMenu->nitem - 1, -1);
+	goto_menu(CurrentMenu, CurrentMenu->keyselect[(int)c] * nlmKeys, 1);
+    }
+    return (MENU_NOTHING);
+}
+
+static int
+lmSelect(char c)
+{
+    if (IS_ASCII(c))
+	return select_menu(CurrentMenu, (CurrentMenu->select / nlmKeys) *
+			   nlmKeys + CurrentMenu->keyselect[(int)c]);
+    else
+	return (MENU_NOTHING);
+}
+
+Anchor *
+list_menu(Buffer *buf)
+{
+    Menu menu;
+    AnchorList *al = buf->href;
+    Anchor *a;
+    Anchor **ap;
+    int i, n, nitem = 0, key = -1, two = FALSE;
+    char **label;
+    char *t;
+    unsigned char c;
+
+    if (!al)
+	return NULL;
+    for (i = 0; i < al->nanchor; i++) {
+	a = &al->anchors[i];
+	if (!a->slave)
+	    nitem++;
+    }
+    if (!nitem)
+	return NULL;
+
+    if (nitem >= nlmKeys)
+	two = TRUE;
+    label = New_N(char *, nitem + 1);
+    ap = New_N(Anchor *, nitem);
+    for (i = 0, n = 0; i < al->nanchor; i++) {
+	a = &al->anchors[i];
+	if (!a->slave) {
+	    t = getAnchorText(buf, al, a);
+	    if (!t)
+		t = "";
+	    if (two && n >= nlmKeys2 * nlmKeys)
+		label[n] = Sprintf("  : %s", t)->ptr;
+	    else if (two)
+		label[n] = Sprintf("%c%c: %s", lmKeys2[n / nlmKeys],
+				   lmKeys[n % nlmKeys], t)->ptr;
+	    else
+		label[n] = Sprintf("%c: %s", lmKeys[n], t)->ptr;
+	    ap[n] = a;
+	    n++;
+	}
+    }
+    label[nitem] = NULL;
+
+    new_option_menu(&menu, label, &key, NULL);
+
+    menu.initial = 0;
+    menu.cursorX = buf->cursorX + buf->rootX;
+    menu.cursorY = buf->cursorY + buf->rootY;
+    menu.x = menu.cursorX + FRAME_WIDTH + 1;
+    menu.y = menu.cursorY + 2;
+    for (i = 0; i < 128; i++)
+	menu.keyselect[i] = -1;
+    if (two) {
+	for (i = 0; i < nlmKeys2; i++) {
+	    c = lmKeys2[i];
+	    menu.keymap[(int)c] = lmGoto;
+	    menu.keyselect[(int)c] = i;
+	}
+	for (i = 0; i < nlmKeys; i++) {
+	    c = lmKeys[i];
+	    menu.keymap[(int)c] = lmSelect;
+	    menu.keyselect[(int)c] = i;
+	}
+    }
+    else {
+	for (i = 0; i < nitem; i++) {
+	    c = lmKeys[i];
+	    menu.keymap[(int)c] = mSelect;
+	    menu.keyselect[(int)c] = i;
+	}
+    }
+
+    a = retrieveCurrentAnchor(buf);
+    if (a) {
+	for (i = 0; i < nitem; i++) {
+	    if (a->hseq == ap[i]->hseq) {
+		menu.initial = i;
+		break;
+	    }
+	}
+    }
+
+    popup_menu(NULL, &menu);
+
+    return (key >= 0) ? ap[key] : NULL;
+}
+
 #endif				/* USE_MENU */
