@@ -1,9 +1,11 @@
+/* $Id: terms.h,v 1.8 2002/11/15 15:44:36 ukai Exp $ */
 #ifndef TERMS_H
 #define TERMS_H
 
 extern int LINES, COLS;
-
-extern char DisplayCode;
+#if defined(__CYGWIN__) && LANG == JA
+extern int LASTLINE;
+#endif
 
 #define CODE_ASCII	'\0'
 #define CODE_EUC	'E'
@@ -25,14 +27,29 @@ extern char DisplayCode;
 #define STR_JIS_J	"ISO-2022-JP (JIS C 6226 + '\033(H')"
 #define STR_INNER_EUC	"EUC-JP (internal)"
 
-#if defined(__EMX__)&&!defined(JP_CHARSET)
-/*
- * Following definitions are valid only for the OS/2 native console
- */
-#define CODE_850	'8'	/* code page 850	*/
-#define CODE_PC		'P'	/* another code pages	*/
-#endif			/* __EMX__ */
-
 #define CODE_JIS(x) ((x)==CODE_JIS_n||(x)==CODE_JIS_m||(x)==CODE_JIS_N||(x)==CODE_JIS_j||(x)==CODE_JIS_J)
 
+#ifdef USE_MOUSE
+/* Addition:mouse event */
+#define MOUSE_BTN1_DOWN 0
+#define MOUSE_BTN2_DOWN 1
+#define MOUSE_BTN3_DOWN 2
+#define MOUSE_BTN4_DOWN_RXVT 3
+#define MOUSE_BTN5_DOWN_RXVT 4
+#define MOUSE_BTN4_DOWN_XTERM 64
+#define MOUSE_BTN5_DOWN_XTERM 65
+#define MOUSE_BTN_UP 3
+#define MOUSE_BTN_RESET -1
+#endif
+
+#ifdef __CYGWIN__
+#ifdef USE_MOUSE
+extern int cygwin_mouse_btn_swapped;
+#endif
+#ifdef SUPPORT_WIN9X_CONSOLE_MBCS
+extern void enable_win9x_console_input(void);
+extern void disable_win9x_console_input(void);
+#endif
+#endif
+
 #endif				/* not TERMS_H */
